# -*- coding: utf-8 -*-
"""
Compatibility shims.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import array
import functools
import os
import platform
import sys

import BTrees
# XXX: This is a private module in ZODB, but it has a lot
# of knowledge about how to choose the right implementation
# based on Python version and implementation. We at least
# centralize the import from here.
from ZODB._compat import HIGHEST_PROTOCOL
from ZODB._compat import Pickler
from ZODB._compat import Unpickler
from ZODB._compat import dump
from ZODB._compat import dumps
from ZODB._compat import loads

__all__ = [
    # ZODB exports
    'HIGHEST_PROTOCOL',
    'Pickler',
    'Unpickler',
    'dump',
    'dumps',
    'loads',

    # Constants
    'PY3',
    'PY2',
    'PY36',
    'PYPY',
    'WIN',
    'MAC',
    'IN_TESTRUNNER',

    # dicts
    'list_values',
    'iteritems',
    'iterkeys',
    'itervalues',

    # OID and TID datastructures and algorithms
    "OID_TID_MAP_TYPE",
    'OID_OBJECT_MAP_TYPE',
    'OID_SET_TYPE',
    'OidTMap_difference',
    'OidTMap_multiunion',
    'OidTMap_intersection',
    'OidList',

    'MAX_TID',
    'iteroiditems',
    'string_types',
    'NStringIO',
    'metricmethod',
    'metricmethod_sampled',
    'wraps',
    'ABC',
    'base64_encodebytes',
    'base64_decodebytes',
    'update_wrapper',

    # Clocks
    'perf_counter',
]

PY3 = sys.version_info[0] == 3
PY36 = sys.version_info[:2] >= (3, 6)
PY2 = not PY3
PYPY = platform.python_implementation() == 'PyPy'
WIN = sys.platform.startswith('win')
MAC = sys.platform.startswith('darwin')

try:
    # Python 3.3+ (PEP 418)
    from time import perf_counter
except ImportError:
    import time

    if sys.platform == "win32":
        perf_counter = time.clock # pylint: disable=no-member
    else:
        perf_counter = time.time

    del time

# Dict support

if PY3:
    def list_values(d):
        return list(d.values())
    iteritems = dict.items
    iterkeys = dict.keys
    itervalues = dict.values
else:
    list_values = dict.values
    iteritems = dict.iteritems  # pylint:disable=no-member
    iterkeys = dict.iterkeys  # pylint:disable=no-member
    itervalues = dict.itervalues  # pylint:disable=no-member

# OID and TID data structures.
#
# The cache MVCC implementation depends on the map types being atomic
# for primitive operations, so don't accept Python BTree
# implementations. (Also, on PyPy, the Python BTree implementation
# uses more memory than a dict.)
#
# The map types are rarely large, but they could sometimes be.
# The set types are more likely to be large (e.g., a set is used to hold
# all the OIDs we need to poll after reading from a persistent cache).

if BTrees.LLBTree.LLBTree is not BTrees.LLBTree.LLBTreePy: # pylint:disable=no-member
    # For BTree and Tree set objects, if you subclass you can define two attributes
    # to customize their allocation (simplified):
    #
    # - ``max_internal_size`` is the number of buckets the root BTree
    # is allowed to hold before it splits. (Actually, it goes to twice
    # that size before it splits.) Each time it needs to resize its
    # internal array, it doubles. In the C code, the default for this
    # is DEFAULT_MAX_BTREE_SIZE == 500.
    #
    # - ``max_leaf_size`` is how big (strictly) a bucket is allowed to
    # be before *it* splits. Buckets start out by allocating an array
    # of size 16 and they double their allocation each time they have
    # to grow. Unfortunately there's no way to give them a suggested
    # size and have them allocate that. In the C code, the default for
    # this is DEFAULT_MAX_BUCKET_SIZE == 120
    #
    # Since these are only in-memory, there's no real downside to
    # increasing them all substantially; it will result in larger
    # memory allocations and better memory locality. The algorithms
    # are still logarithmic (and actually creation and lookups
    # benchmark slightly faster for larger values). The only time that
    # could be a problem is if they're *very* large, then the amount
    # of waste could be substantial. Choose values one less than a
    # power of two to avoid that last huge allocation.
    #
    # Using these values and an OidSet of about 800,000 objects (from
    # a real cache file) results in 12,000 fewer LLSet objects being
    # created, and 2MB less memory being used.
    class OID_TID_MAP_TYPE(BTrees.family64.UU.BTree):
        __slots__ = ()
        # Since these are contiguous arrays, there *might* be some
        # benefit in keeping them within a virtual memory page
        # (commonly 4K). The BTree doesn't actually hold an array of
        # Bucket, it holds an array of BTreeItem, which is a key and a
        # pointer (to bucket or tree). Assuming 64-bit pointers, the size of an array is
        # (8 * 2 * len) and we'd have to go down to 255, which is actually a
        # reduction from the default of 500 (8000 bytes, less than two pages).
        # Until we have evidence that keeping in a page matters, we'll just choose
        # some walues that we observed to result in an overall memory use reduction.
        max_internal_size = 1023 # just under 16K, four pages.
        # buckets have an array of keys and (for trees but not sets) an array of values.
        # This is definitely 64 bits
        max_leaf_size = 4095 # under 32K, or 8 pages.

    class OID_OBJECT_MAP_TYPE(BTrees.family64.UO.BTree):
        __slots__ = ()
        max_internal_size = OID_TID_MAP_TYPE.max_internal_size
        max_leaf_size = OID_TID_MAP_TYPE.max_leaf_size

    class OID_SET_TYPE(BTrees.family64.UU.TreeSet):
        max_internal_size = OID_TID_MAP_TYPE.max_internal_size
        max_leaf_size = OID_TID_MAP_TYPE.max_leaf_size

    OidTMap_difference = BTrees.family64.UU.difference  # pylint:disable=no-member
    OidTMap_multiunion = BTrees.family64.UU.multiunion  # pylint:disable=no-member
    OidTMap_intersection = BTrees.family64.UU.intersection  # pylint:disable=no-member
    OidSet_difference = OidTMap_difference
    def OidSet_discard(s, val):
        try:
            s.remove(val)
        except KeyError:
            pass

    def OidObjectMap_max_key(bt):
        if not bt:
            return 0
        return bt.maxKey()
else:
    OID_TID_MAP_TYPE = dict
    OID_OBJECT_MAP_TYPE = dict
    OID_SET_TYPE = set
    def OidTMap_difference(c1, c2):
        # Must prevent iterating while being changed
        c1 = dict(c1)
        return {k: c1[k] for k in set(c1) - set(c2)}

    def OidTMap_multiunion(seq):
        return set().union(*seq)

    def OidTMap_intersection(c1, c2):
        return set(c1).intersection(set(c2))

    def OidSet_difference(c1, c2):
        return set(c1) - set(c2)

    OidSet_discard = set.discard

    def OidObjectMap_max_key(mapping):
        if not mapping:
            return 0
        return max(iterkeys(mapping))

# Lists of OIDs or TIDs. These could be simple list() objects, or we
# can treat them as numbers and store them in array.array objects, if
# we have an unsigned 64-bit element type. array.array, just like the
# C version of BTrees, uses less memory on CPython, but has a cost
# converting back and forth between objects and native values. What's
# the cost? Let's measure.
#
# Test: list(xrange(30000000)) vs array.array('L', xrange(30000000))
#  on Python 2, with minor modifications (range and 'Q') on Python 3.
#
#              list mem  | array mem | list time | array time
# CPython 2:      861MB  |     228MB |    596ms  |     2390ms
# PyPy2 7.1:      229MB  |     227MB |    178ms  |     1830ms
# CPython 3.7:   2117MB  |     232MB |   3680ms  |     3150ms
#
# Test: Same as above, but using 300 instead of 30000000
#               list time | array time
# CPython 2:       6.28ms |     6.3ms
# PyPy2 7.1:       1.34ms |     1.43ms
# CPython 3.7:     3.69ms |     3.74ms
#
# Slicing x(30000000)[30000:30200]
#               list time | array time
# CPython 2:       427ns  |      148ns
# PyPy2 7.1*:      138ns  |     8950ns
# CPython 3.7:     671ns  |      411ns
#
# iterate x(30000000): for _ in x: pass
#               list time | array time  | small list time | small array time
# CPython 2:       357ms  |      604ms  |    2640ns       |  6050ns
# PyPy2 7.1*:       51ms  |      592ms  |     601ns       |  5910ns
# CPython 3.7:     308ms  |     2240ms  |    2250ns       |  6170ns
# * On PyPy, the test was wrapped in a method for better JIT.
#
# Using BTrees.family64.II.TreeSet(range(30000000))
#
#                memory  | construction time | iteration time
# CPython 2:      564MB  |            2740ms |    520ms
# CPython 3.7:    573MB  |            5280ms |   2390ms
#
#
# Observations:
# - Large list() is faster to create on CPython 2, but uses 4x the memory.
# - Large list() is *slower* to create on CPython 3 and uses an incredible
#    9x the memory. Relative to Python 2, I suspect the differences have to do with
#    all Python 3 integers being variable-length long objects, unlike Python 2.
#    I suspect that accounts for much of the difference in general.
# - PyPy memory usage is comparable for both list and array (which makes sense, it has
#    a specialized strategy for lists of integers), but large lists are faster to
#    create for some reason.
# - Creation times for small sets is basically the same on all platforms.
# - Slicing time of arrays is faster on CPython 2 and 3 but much slower on PyPy.
# - Iterating arrays is substantially slower on all platforms and for all sizes.
# - However, creating arrays is faster than creating 64-bit TreeSets; iteration
#   is about the same.
#
# Conclusions:
# Except on PyPy, when working with a large list of OIDs, a 64-bit array.array
# will save a substantial amount of memory. On Python 3, it will probably be slightly
# faster to create too; on both Python 2 and 3 it will be faster and smaller than an equivalent
# TreeSet. Slicing is faster with arrays as well. Iteration is around 3x slower, but that's likely
# to be noise compared to the body of the loop.
# Thus, everywhere except PyPy, if we have an unsigned 64-bit array.array available, that should
# be our choice.
_64bit_array = None
try:
    # Find out if we have a native unsigned 64-bit type
    array.array('Q', [1])
    _64bit_array = functools.partial(array.array, 'Q')
except ValueError:
    # We don't. Either we're on Python 2 or the compiler doesn't support 'long long'.
    # What about a regular unsigned long? If we're on a 64-bit platform, that
    # might be enough.
    a = array.array('L', [1])
    if a.itemsize >= 8:
        _64bit_array = functools.partial(array.array, 'L')

if _64bit_array and not PYPY:
    OidList = _64bit_array
else:
    OidList = list
TidList = OidList
# The maximum theoretical TID (and OID, incidentally).
# This is b"\xff\xff\xff\xff\xff\xff\xff\xff"
# It requires an 8 byte unsigned integer to store.
MAX_TID = BTrees.family64.maxuint
# The maximum TID if only signed values are allowed.
MAX_S_TID = BTrees.family64.maxint

def iteroiditems(d):
    # Could be either a BTree, which always has 'iteritems',
    # or a plain dict, which may or may not have iteritems.
    return d.iteritems() if hasattr(d, 'iteritems') else d.items()

# Types
from perfmetrics import metricmethod # pylint:disable=wrong-import-position
from perfmetrics import Metric # pylint:disable=wrong-import-position

if PY3:
    string_types = (str,)
    number_types = (int, float)
    from io import StringIO as NStringIO
    from functools import wraps
else:
    string_types = (basestring,) # pylint:disable=undefined-variable
    number_types = (int, long, float) # pylint:disable=undefined-variable
    from io import BytesIO as NStringIO
    # On Python 2, functools.update_wrapper doesn't set the '__wrapped__'
    # attribute, and we need that.
    from functools import wraps as _wraps
    class wraps(object):
        def __init__(self, func):
            self._orig = func
            self._wrapper = _wraps(func)

        def __call__(self, replacement):
            replacement = self._wrapper(replacement)
            replacement.__wrapped__ = self._orig
            return replacement

metricmethod_sampled = Metric(method=True, rate=0.1)

IN_TESTRUNNER = (
    # zope-testrunner --test-path ...
    'zope-testrunner' in sys.argv[0]
    # python -m zope.testrunner --test-path ...
    or os.path.join('zope', 'testrunner') in sys.argv[0]
)


if IN_TESTRUNNER and os.environ.get('RS_TEST_DISABLE_METRICS'):
    # If we're running under the testrunner,
    # don't apply the metricmethod stuff. It makes
    # backtraces ugly and makes stepping in the
    # debugger annoying.
    metricmethod = metricmethod_sampled = lambda f: f

try:
    from abc import ABC
except ImportError:
    import abc
    ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})
    del abc

# Functions
if PY3:
    xrange = range
    intern = sys.intern
    from base64 import encodebytes as base64_encodebytes
    from base64 import decodebytes as base64_decodebytes
    casefold = str.casefold
    from traceback import clear_frames
    clear_frames = clear_frames # pylint:disable=self-assigning-variable
    from functools import update_wrapper
else:
    xrange = xrange # pylint:disable=self-assigning-variable
    intern = intern # pylint:disable=self-assigning-variable
    from base64 import encodestring as base64_encodebytes
    from base64 import decodestring as base64_decodebytes
    casefold = str.lower
    def clear_frames(tb): # pylint:disable=unused-argument
        "Does nothing on Py2."

    from functools import update_wrapper as _update_wrapper
    def update_wrapper(wrapper, wrapped, *args, **kwargs):
        wrapper = _update_wrapper(wrapper, wrapped, *args, **kwargs)
        wrapper.__wrapped__ = wrapped
        return wrapped



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
MySQL Connector/Python IDBDriver implementations.
"""
from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from relstorage._compat import PY2
from relstorage._compat import PYPY
from relstorage.adapters.interfaces import IDBDriver

from . import AbstractMySQLDriver

__all__ = [
    'PyMySQLConnectorDriver',
    'CMySQLConnectorDriver',
]

logger = __import__('logging').getLogger(__name__)

_base_name = 'MySQL Connector/Python'


@implementer(IDBDriver)
class PyMySQLConnectorDriver(AbstractMySQLDriver):
    # See https://github.com/zodb/relstorage/issues/155
    __name__ = 'Py ' + _base_name

    MODULE_NAME = 'mysql.connector'
    PRIORITY = 4
    PRIORITY_PYPY = 2

    USE_PURE = True

    _CONVERTER_CLASS = None
    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = True

    def __init__(self):
        super(PyMySQLConnectorDriver, self).__init__()
        # This driver doesn't support ``init_command``, we have to run
        # it manually.
        self.__init_command = self._init_command
        del self._init_command

        # conn.close() -> InternalError: Unread result found
        # By the time we get to a close(), it's too late to do anything about it.
        self.close_exceptions += (self.driver_module.InternalError,)
        if self.Binary is str:
            self.Binary = bytearray

        if PYPY:
            # Patch to work around JIT bug found in (at least) 7.1.1
            # https://bitbucket.org/pypy/pypy/issues/3014/jit-issue-inlining-structunpack-hh
            #
            # In addition, we later discovered
            # https://github.com/zodb/relstorage/issues/283#issuecomment-516489791,
            # which was tracked down to be the same thing: protocol calls into utils
            # to parse "length code" for lastrowid in utils.read_lc_int, which uses
            # struct_unpack
            try:
                from mysql.connector import catch23
                from mysql.connector import protocol
                from mysql.connector import utils
            except ImportError: # pragma: no cover
                catch23 = protocol = utils = None

            if not hasattr(catch23, 'struct_unpack') or not hasattr(protocol, 'struct_unpack'):
                # pragma: no cover
                logger.debug("Unknown mysql.connector; not patching for PyPy JIT")
            else:
                logger.debug("Patching mysql.connector for PyPy JIT bug")
                from struct import unpack
                def struct_unpack(fmt, buf):
                    if isinstance(buf, bytearray):
                        buf = bytes(buf)
                    return unpack(fmt, buf)

                catch23.struct_unpack = struct_unpack
                protocol.struct_unpack = struct_unpack
                utils.struct_unpack = struct_unpack

    @classmethod
    def _get_converter_class(cls):
        if cls._CONVERTER_CLASS is None:
            import mysql.connector.conversion # pylint:disable=import-error

            # The results of things like 'SHOW TABLES' come in
            # and use _STRING_to_python (or _VAR_STRING_to_python).
            # This uses the value of self._use_unicode to decide whether to
            # decode bytes and bytearrays to unicode, unless the charset is
            # set to binary. Unfortunately, there's a bug and it doesn't
            # ever actually read 'use_unicode' out of the keyword arguments!
            # We want strings to be native strings on Python 2.

            # It likes to return BLOB columns as Unicode strings too! Clearly
            # that's not right. (The reason is because apparently JSON data is
            # stored with the same type as a BLOB...so for every BLOB it sees, it
            # first tries to parse it as JSON!)

            # The Python implementation calls row_to_python(), the C version
            # calls to_python()
            class BlobConverter(mysql.connector.conversion.MySQLConverter):

                def __init__(self, charset_name, use_unicode):
                    if PY2:
                        use_unicode = False
                    super(BlobConverter, self).__init__(charset_name, use_unicode)

                if PY2:
                    def _STRING_to_python(self, value, dsc=None):
                        # This is also supposed to handle SET, but we don't use that type
                        # anywhere
                        if isinstance(value, bytearray):
                            # The Python version tends to return bytearray values.
                            return bytes(value)
                        # The C extension tends to return byte (aka str) values.
                        return super(BlobConverter, self)._STRING_to_python(value, dsc)

                    _VAR_STRING_to_python = _STRING_to_python

                # There are a few places we get into trouble on
                # Python 2/3 with bytearrays comping back: they
                # can't be hashed for the local_client compression
                # functions or sent to zlib.decompress(), they
                # can't be sent to pickle.loads(), etc, so it's
                # best to return them as bytes.
                def _BLOB_to_python(self, value, dsc=None): # pylint:disable=unused-argument
                    if isinstance(value, bytearray):
                        return bytes(value)
                    return value or b''

                _LONG_BLOB_to_python = _BLOB_to_python
                _MEDIUM_BLOB_to_python = _BLOB_to_python
                _TINY_BLOB_to_python = _BLOB_to_python


            cls._CONVERTER_CLASS = BlobConverter

        return cls._CONVERTER_CLASS

    def connect(self, *args, **kwargs):
        # It defaults to the (slower) pure-python version prior to 8.0.11.
        # NOTE: The C implementation doesn't support the prepared
        # operations.
        # NOTE: The C implementation returns bytes when the Py implementation
        # returns bytearray under Py2
        kwargs['use_pure'] = self.USE_PURE
        kwargs['converter_class'] = self._get_converter_class()
        kwargs['get_warnings'] = True
        # By default, make it fetch all rows for the cursor, like most
        # drivers do. Unless we do this, cursors won't buffer, so we
        # don't know how many rows there are. That's fine and within
        # the DB-API spec. The Python implementation is much faster if
        # we don't ask it to. The C connection doesn't accept the
        # 'prepared' keyword. You can't have both a buffered and
        # prepared cursor, but the prepared cursor doesn't gain us
        # anything anyway.
        kwargs['buffered'] = True
        conn = super(PyMySQLConnectorDriver, self).connect(*args, **kwargs)
        cur = conn.cursor()
        try:
            cur.execute(self.__init_command)
        finally:
            cur.close()
        return conn

    def cursor(self, conn, server_side=False):
        if server_side:
            cursor = conn.cursor(buffered=False)
            cursor.arraysize = self.cursor_arraysize
        else:
            cursor = super(PyMySQLConnectorDriver, self).cursor(conn, server_side=server_side)
        cursor.connection = conn
        return cursor

    def set_autocommit(self, conn, value):
        # This implementation uses a property instead of a method.
        conn.autocommit = value

    def callproc_multi_result(self, cursor, proc, args=(), exit_critical_phase=False):
        # This driver is weird, wants multi=True, returns an iterator of cursors
        # instead of using nextset()
        resultsets = cursor.execute("CALL " + proc, args, multi=True)
        multi_results = []
        for resultset in resultsets:
            try:
                multi_results.append(resultset.fetchall())
            except self.driver_module.InterfaceError:
                # This gets raised on the empty set at the end, for some reason.
                # Ensure we put one there to be like the others
                multi_results.append(())
                break
        return multi_results

    def callproc_no_result(self, cursor, proc, args=()):
        # Again, weird. The call's empty result set seems to be consumed.
        cursor.execute("CALL " + proc, args)

class CMySQLConnectorDriver(PyMySQLConnectorDriver):
    __name__ = 'C ' + _base_name

    AVAILABLE_ON_PYPY = False
    # Values this high won't usually be automatically tested.
    # This driver (as of 8.0.16) fails to run under Python 3.7's development
    # mode, with assertion failures about doing things without the GIL
    # that crash the interpreter. That's pretty risky, so we don't currently
    # recommend it.
    # TODO: 8.0.17 claims to have fixed issues with Python 3. Verify this.
    PRIORITY = 1000
    PRIORITY_PYPY = 4000

    USE_PURE = False
    _GEVENT_CAPABLE = False

    def get_driver_module(self):
        mod = super(CMySQLConnectorDriver, self).get_driver_module()
        if not mod.HAVE_CEXT:
            raise ImportError("No C extension")
        return mod

    _C_CONVERTER_CLASS = None

    @classmethod
    def _get_converter_class(cls):
        # The C implementation does *some* of its conversion down in the C
        # layer. If we try to pass that through the conversion process again,
        # we get bugs.
        # Notably, integer columns that have a value of 0 are interpreted as
        # NULL and end processing of the result set! This means, for example, that
        # we can't get changes for the root object, because it has zoid == 0.
        # We workaround that here.

        if cls._C_CONVERTER_CLASS is None:
            from mysql.connector.conversion import FieldType # pylint:disable=import-error
            # pylint:disable=inherit-non-class
            class Converter(PyMySQLConnectorDriver._get_converter_class()):
                def to_python(self, vtype, value):
                    if value == 0 and vtype[1] == FieldType.LONGLONG:
                        return value
                    return super(Converter, self).to_python(vtype, value)

            cls._C_CONVERTER_CLASS = Converter

        return cls._C_CONVERTER_CLASS



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""IObjectMover implementation.
"""
from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from relstorage.adapters.interfaces import IObjectMover

from ..schema import Schema
from ..mover import AbstractObjectMover
from ..mover import metricmethod_sampled


@implementer(IObjectMover)
class MySQLObjectMover(AbstractObjectMover):

    _create_temp_store = Schema.temp_store.create()

    @metricmethod_sampled
    def on_store_opened(self, cursor, restart=False):
        """Create the temporary table for storing objects"""
        if restart:
            # TRUNCATE is a DDL statement, even against a temporary
            # table, and as such does an implicit transaction commit.
            # Normally we want to avoid that, but here its OK since
            # this method is called between transactions, as it were.
            # TRUNCATE benchmarks (zodbshoot add) substantially faster
            # (10157) than a DELETE (75xx) and moderately faster than
            # a DROP/CREATE (9457). TRUNCATE is in the replication
            # logs like a DROP/CREATE. (DROP TEMPORARY TABLE is *not*
            # DDL and not transaction ending).
            # We are up to 4 temp tables, doing this with a call to a stored
            # proc saves round trips.
            #
            # It's possible that the DDL lock that TRUNCATE takes can be a bottleneck
            # in some places, though?

            self.driver.callproc_no_result(
                cursor,
                "clean_temp_state(false)"
            )
        else:
            # InnoDB tables benchmark much faster for concurrency=2
            # and 6 than MyISAM tables under both MySQL 5.5 and 5.7,
            # at least on OS X 10.12. The OS X filesystem is single
            # threaded, though, so the effects of flushing MyISAM tables
            # to disk on every operation are probably magnified.

            # note that the md5 column is not used if self.keep_history == False.
            self._create_temp_store.execute(cursor)

            stmt = """
            CREATE TEMPORARY TABLE temp_read_current (
                zoid        BIGINT UNSIGNED NOT NULL PRIMARY KEY,
                tid         BIGINT UNSIGNED NOT NULL
            ) ENGINE InnoDB
            """
            cursor.execute(stmt)

            stmt = """
            CREATE TEMPORARY TABLE temp_blob_chunk (
                zoid        BIGINT UNSIGNED NOT NULL,
                chunk_num   BIGINT UNSIGNED NOT NULL,
                            PRIMARY KEY (zoid, chunk_num),
                chunk       LONGBLOB
            ) ENGINE InnoDB
            """
            cursor.execute(stmt)

            stmt = """
            CREATE TEMPORARY TABLE IF NOT EXISTS temp_locked_zoid (
                zoid BIGINT UNSIGNED NOT NULL PRIMARY KEY
            ) ENGINE InnoDB
            """
            cursor.execute(stmt)

        AbstractObjectMover.on_store_opened(self, cursor, restart=restart)

    @metricmethod_sampled
    def restore(self, cursor, batcher, oid, tid, data):
        """Store an object directly, without conflict detection.

        Used for copying transactions into this database.
        """
        if self.keep_history:
            suffix = """
            ON DUPLICATE KEY UPDATE
                tid = VALUES(tid),
                prev_tid = VALUES(prev_tid),
                md5 = VALUES(md5),
                state_size = VALUES(state_size),
                state = VALUES(state)
            """
        else:
            suffix = """
            ON DUPLICATE KEY UPDATE
                tid = VALUES(tid),
                state_size = VALUES(state_size),
                state = VALUES(state)
            """
        self._generic_restore(batcher, oid, tid, data,
                              command='INSERT', suffix=suffix)

    # Override this query from the superclass. The MySQL optimizer, up
    # through at least 5.7.17 doesn't like actual subqueries in a DELETE
    # statement. See https://github.com/zodb/relstorage/issues/175
    _move_from_temp_hf_delete_blob_chunk_query = """
    DELETE bc
    FROM blob_chunk bc
    INNER JOIN (SELECT zoid FROM temp_store) sq
    ON bc.zoid = sq.zoid
    """

    # We UPSERT for hf movement; no need to do a delete.
    _move_from_temp_hf_delete_query = ''



# -*- coding: utf-8 -*-
"""
The Oracle dialect

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from contextlib import contextmanager

from ..sql import DefaultDialect
from ..sql import Compiler
from ..sql import Boolean
from ..sql import Column
from ..sql import OID
from ..sql import TID

from ..sql._util import Columns
from ..sql.insert import _ExcludedColumn

class OracleCompiler(Compiler):

    def visit_boolean_literal_expression(self, value):
        sql = "'Y'" if value else "'N'"
        self.emit(sql)

    def can_prepare(self):
        # We haven't investigated preparing statements manually
        # with cx_Oracle. There's a chance that `cx_Oracle.Connection.stmtcachesize`
        # will accomplish all we need.
        return False

    def _placeholder(self, key):
        # XXX: What's that block in the parent about key == '%s'?
        # Only during prepare() I think.
        return ':' + key

    def visit_ordered_bind_param(self, bind_param):
        ph = self.placeholders[bind_param] = ':%d' % (len(self.placeholders) + 1,)
        self.emit(ph)

    def visit_select_expression(self, column_node):
        if isinstance(column_node, Column) and column_node.is_type(Boolean):
            # Fancy CASE statement to get 1 or 0 into Python
            self.emit_keyword('CASE WHEN')
            self.emit_identifier(column_node.name)
            self.emit(" = 'Y' THEN 1 ELSE 0 END")
        else:
            super(OracleCompiler, self).visit_select_expression(column_node)

    @contextmanager
    def visit_limited_select(self, select, limit):
        if limit:
            self.visit_limit = self._visit_limit_while_limited
            self.emit('SELECT * FROM (')
            try:
                yield self
            finally:
                del self.visit_limit
            self.emit(') WHERE ROWNUM <=')
            self.visit_immediate_expression(limit)
        else:
            yield self

    def visit_limit(self, limit_literal): # pylint:disable=method-hidden
        if limit_literal is not None:
            raise Exception("Forgot to visit in a visit_limited_select")

    def _visit_limit_while_limited(self, limit):
        assert limit is not None

    def visit_upsert(self, upsert):
        self.emit_keyword('MERGE INTO')
        self.visit(upsert.table)
        self.emit_keyword('USING')
        if upsert.select:
            self.visit_grouped(upsert.select)
            self.emit_keyword('D')
        else:
            raise TypeError
        self.emit_keyword('ON')
        self.emit(
            '(',
            upsert.table.name + '.' + upsert.conflict_column.name,
            ' = '
            'D.' + upsert.conflict_column.name,
            ')'
        )
        self.emit_keyword('WHEN MATCHED THEN')
        self.visit(upsert.update_clause)

        self.emit_keyword('WHEN NOT MATCHED THEN INSERT')
        self.visit_grouped(upsert.column_list)
        self.emit_keyword('VALUES')
        excluded_columns = Columns(_ExcludedColumn(c.name) for c in upsert.column_list)
        self.visit_grouped(excluded_columns)

    def visit_upsert_excluded_column(self, column):
        self.emit('D.' + column.name)

class OracleDialect(DefaultDialect):

    datatype_map = DefaultDialect.datatype_map.copy()
    datatype_map.update({
        OID: 'NUMBER(20)',
        TID: 'NUMBER(20)',
        Boolean: "CHAR",
    })

    def extra_constraints_for_column(self, column):
        if column.is_type(Boolean):
            return "CHECK ({col_name} IN ('N', 'Y'))".format(
                col_name=column.name
            )
        return ''

    def compiler_class(self):
        return OracleCompiler



# -*- coding: utf-8 -*-
"""
Tests for oidallocator.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from ..oidallocator import AbstractRangedOIDAllocator


class MockOIDAllocator(AbstractRangedOIDAllocator):

    def _set_min_oid_from_range(self, cursor, n):
        raise NotImplementedError

    def new_oids(self, cursor):
        raise NotImplementedError

class TestOIDAllocator(TestCase):

    alloc = MockOIDAllocator()

    def test_range_around1(self, meth='_oid_range_around', convert=lambda x: x):
        # By default we get a list, no need to convert
        meth = getattr(self.alloc, meth) if not callable(meth) else meth
        oids = meth(1)
        oids = convert(oids)
        self.assertEqual(
            oids,
            [16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1])

    def test_range_around1_list(self):
        # Range may or may not actually return a list, depending on the version,
        # so we convert
        from ..oidallocator import _oid_range_around_assume_list
        self.test_range_around1(_oid_range_around_assume_list, list)

    def test_range_around1_iterable(self):
        from ..oidallocator import _oid_range_around_iterable
        self.test_range_around1(_oid_range_around_iterable)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.tests import MockConnection
from relstorage.tests import MockConnectionManager
from relstorage.tests import MockCursor
from relstorage.tests import MockPoller

from ..interfaces import ITransactionControl

class MockStoreConnection(object):

    def __init__(self, conn):
        self.connection = conn

    def rollback_quietly(self):
        self.connection.rollback()

    def commit(self):
        self.connection.commit()

class TestTransactionControl(TestCase):

    def _getClass(self):
        from ..txncontrol import GenericTransactionControl
        return GenericTransactionControl

    def Binary(self, arg):
        if not isinstance(arg, bytes):
            arg = arg.encode('ascii')
        return arg

    def _makeOne(self, keep_history=True, binary=None):
        return self._getClass()(MockConnectionManager(), MockPoller(),
                                keep_history, binary or self.Binary)

    def test_provides(self):
        assert_that(self._makeOne(), validly_provides(ITransactionControl))

    def test_get_tid_empty_db(self):
        inst = self._makeOne()
        inst.poller.poll_tid = 0
        cur = MockCursor()
        cur.results = None

        self.assertEqual(inst.get_tid(cur), 0)

    def test_add_transaction_hp(self):
        inst = self._makeOne()
        cur = MockCursor(self)
        __traceback_info__ = inst.__dict__
        inst.add_transaction(cur, 1, u'user', u'desc', u'ext')
        self.assertEqual(
            cur.executed.pop(),
            (str(inst._add_transaction_query),
             (1, False, b'user', b'desc', b'ext'))
        )

        inst.add_transaction(cur, 1, u'user', u'desc', u'ext', packed=True)

        self.assertEqual(
            cur.executed.pop(),
            (str(inst._add_transaction_query),
             (1, True, b'user', b'desc', b'ext'))
        )

    def test_commit_phase1(self):
        inst = self._makeOne()
        result = inst.commit_phase1(None, None)
        self.assertEqual(result, '-')

    def test_commit_phase2(self):
        inst = self._makeOne()
        conn = MockConnection()
        inst.commit_phase2(MockStoreConnection(conn), None, None)
        self.assertTrue(conn.committed)

    def test_abort(self):
        inst = self._makeOne()
        conn = MockConnection()
        inst.abort(MockStoreConnection(conn), None)
        self.assertTrue(conn.rolled_back)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for expressions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from .. import expressions

class TestBinaryExpression(TestCase):

    def test_str(self):

        exp = expressions.BinaryExpression('=', 'lhs', 'rhs')
        self.assertEqual(
            str(exp),
            'lhs = rhs'
        )


class TestEmptyExpression(TestCase):

    exp = expressions.EmptyExpression()

    def test_boolean(self):
        self.assertFalse(self.exp)

    def test_str(self):
        self.assertEqual(str(self.exp), '')

    def test_and(self):
        self.assertIs(self.exp.and_(self), self)


class TestAnd(TestCase):

    def test_resolve(self):
        # This will wrap them in literal nodes, which
        # do nothing when resolved.

        exp = expressions.And('a', 'b')

        resolved = exp.resolve_against(None)

        self.assertIsInstance(resolved, expressions.And)
        self.assertIs(resolved.lhs, exp.lhs)
        self.assertIs(resolved.rhs, exp.rhs)



# -*- coding: utf-8; -*-
##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A foundation for RelStorage tests"""
from __future__ import absolute_import
from __future__ import print_function

# pylint:disable=too-many-ancestors,abstract-method,too-many-public-methods,too-many-lines
# pylint:disable=too-many-statements,too-many-locals
import contextlib
import functools
import os
import random
import shutil
import tempfile
import time
import threading
import unittest

import transaction
from persistent import Persistent
from persistent.mapping import PersistentMapping
from zc.zlibstorage import ZlibStorage

import ZODB.tests.util

from ZODB.Connection import TransactionMetaData
from ZODB.DB import DB
from ZODB.FileStorage import FileStorage
from ZODB.POSException import ReadConflictError
from ZODB.POSException import ReadOnlyError
from ZODB.serialize import referencesf
from ZODB.utils import z64
from ZODB.utils import u64 as bytes8_to_int64
from ZODB.utils import p64 as int64_to_8bytes

from ZODB.tests import BasicStorage
from ZODB.tests import ConflictResolution
from ZODB.tests import MTStorage
from ZODB.tests import PackableStorage
from ZODB.tests import PersistentStorage
from ZODB.tests import ReadOnlyStorage
from ZODB.tests import StorageTestBase
from ZODB.tests import Synchronization
from ZODB.tests.StorageTestBase import zodb_pickle
from ZODB.tests.StorageTestBase import zodb_unpickle
from ZODB.tests.MinPO import MinPO

from . import fakecache
from . import util
from . import mock
from . import TestCase
from . import StorageCreatingMixin
from . import skipIfNoConcurrentWriters
from .persistentcache import PersistentCacheStorageTests
from .locking import TestLocking
from .test_zodbconvert import FSZODBConvertTests

class RelStorageTestBase(StorageCreatingMixin,
                         TestCase,
                         StorageTestBase.StorageTestBase):

    base_dbname = None # Override
    keep_history = None  # Override
    _storage_created = None

    def _close(self):
        # Override from StorageTestBase.

        # Try to avoid creating one through our _storage property.
        if '_storage' in self.__dict__:
            storage = self._storage
        else:
            storage = self._storage_created
        self._storage = None

        if storage is not None:
            storage.close()
            storage.cleanup()

    def make_storage_to_cache(self):
        return self.make_storage()

    def get_storage(self):
        # Create a storage with default options
        # if it has not been created already.
        storage = self._storage_created
        if storage is None:
            storage = self.make_storage_to_cache()
            self._storage_created = storage
        return storage

    def set_storage(self, storage):
        self._storage_created = storage

    _storage = property(
        lambda self: self.get_storage(),
        lambda self, nv: self.set_storage(nv)
    )

    def open(self, read_only=False, **kwargs):
        # This is used by a few ZODB tests that close and reopen the storage.
        storage = self._storage
        if storage is not None:
            self._storage = None
            storage.close()
            storage.cleanup()
        self._storage = storage = self.make_storage(
            read_only=read_only, zap=False, **kwargs)
        return storage


class StorageClientThread(MTStorage.StorageClientThread):
    # MTStorage assumes that the storage object is thread safe.
    # This doesn't make any sense for an MVCC Storage like RelStorage;
    # don't try to use a single instance in multiple threads.
    #
    # This patch makes it respect that.

    def __init__(self, storage, *args, **kwargs):
        storage = storage.new_instance()
        super(StorageClientThread, self).__init__(storage, *args, **kwargs)

    def runtest(self):
        try:
            super(StorageClientThread, self).runtest()
        finally:
            self.storage.release()
            self.storage = None


class ExtStorageClientThread(StorageClientThread, MTStorage.ExtStorageClientThread):
    "Same as above."

class ThreadWrapper(object):

    def __init__(self, storage):
        self.__storage = storage
        # We can't use an RLock, which verifies that the thread that
        # acquired is the one that releases; check_tid_ordering_w_commit
        # deliberately spreads these actions across threads (for same reason).
        self.__commit_lock = threading.Lock()
        rl = self.__read_lock = threading.Lock()
        self.__txn = None

        def make_locked(name):
            meth = getattr(storage, name)
            @functools.wraps(meth)
            def func(*args, **kwargs):
                with rl:
                    return meth(*args, **kwargs)
            return func

        for name in (
                'loadBefore',
                'load',
                'store',
                'getTid',
                'lastTransaction',
        ):
            setattr(self, name, make_locked(name))

    def __getattr__(self, name):
        return getattr(self.__storage, name)

    def tpc_begin(self, txn):
        self.__commit_lock.acquire()
        self.__read_lock.acquire()
        assert not self.__txn
        self.__txn = txn
        self.__read_lock.release()
        return self.__storage.tpc_begin(txn)

    def tpc_finish(self, txn, callback=None):
        self.__read_lock.acquire()
        assert txn is self.__txn
        try:
            return self.__storage.tpc_finish(txn, callback)
        finally:
            self.__txn = None
            self.__commit_lock.release()
            self.__read_lock.release()

    def tpc_abort(self, txn):
        self.__read_lock.acquire()
        assert txn is self.__txn, (txn, self.__txn)
        try:
            return self.__storage.tpc_abort(txn)
        finally:
            self.__txn = None
            self.__commit_lock.release()
            self.__read_lock.release()

class UsesThreadsOnASingleStorageMixin(object):
    # These tests attempt to use threads on a single storage object.
    # That doesn't make sense with MVCC, where every instance is its
    # own connection and doesn't need to do any locking. This mixin makes
    # those tests use a special storage that locks.

    @contextlib.contextmanager
    def __thread_safe_wrapper(self):
        orig_storage = self._storage
        wrapped = self._storage = ThreadWrapper(orig_storage)
        try:
            yield
        finally:
            if self._storage is wrapped:
                self._storage = orig_storage

    def __generic_wrapped_test(self, meth_name):
        meth = getattr(
            super(UsesThreadsOnASingleStorageMixin, self),
            meth_name)
        try:
            with self.__thread_safe_wrapper():
                meth()
        finally:
            self._storage.zap_all(slow=True)

    def make_func(name): # pylint:disable=no-self-argument
        return lambda self: self.__generic_wrapped_test(name)

    for bad_test in (
            'check_checkCurrentSerialInTransaction',
            # This one stores a b'y' (invalid pickle) into the
            # database as the root object, so if we don't get zapped
            # afterwards, we can't open the database.
            'check_tid_ordering_w_commit',
    ):
        locals()[bad_test] = make_func(bad_test)

    del make_func
    del bad_test


class GenericRelStorageTests(
        UsesThreadsOnASingleStorageMixin,
        RelStorageTestBase,
        PersistentCacheStorageTests,
        TestLocking,
        BasicStorage.BasicStorage,
        PackableStorage.PackableStorage,
        Synchronization.SynchronizedStorage,
        ConflictResolution.ConflictResolvingStorage,
        PersistentStorage.PersistentStorage,
        MTStorage.MTStorage,
        ReadOnlyStorage.ReadOnlyStorage,
    ):

    def setUp(self):
        # ZODB.tests.util.TestCase likes to change directories
        # It tries to change back in tearDown(), but if there's an error,
        # we may not get to tearDown. addCleanup() always runs, though.
        # do that as the very last thing that happens (except for subclasses, they
        # could add things first)
        self.addCleanup(os.chdir, os.getcwd())

        super(GenericRelStorageTests, self).setUp()
        # PackableStorage is particularly bad about leaving things
        # dangling. For example, if the ClientThread runs into
        # problems, it doesn't close its connection, which can leave
        # locks dangling until GC happens and break other threads and even
        # other tests.
        #
        # Patch around that. Be sure to only close a given connection once,
        # though.
        _closing = self._closing
        def db_factory(storage, *args, **kwargs):
            db = _closing(DB(storage, *args, **kwargs))
            db_open = db.open
            def o(transaction_manager=None, at=None, before=None):
                conn = db_open(transaction_manager=transaction_manager,
                               at=at,
                               before=before)

                _closing(conn)
                if transaction_manager is not None:
                    # If we're using an independent transaction, abort it *before*
                    # attempting to close the connection; that means it must be registered
                    # after the connection.
                    self.addCleanup(transaction_manager.abort)
                return conn
            db.open = o
            return db
        PackableStorage.DB = db_factory

        self.addCleanup(setattr, MTStorage,
                        'StorageClientThread', MTStorage.StorageClientThread)
        MTStorage.StorageClientThread = StorageClientThread

        self.addCleanup(setattr, MTStorage,
                        'ExtStorageClientThread', MTStorage.ExtStorageClientThread)
        MTStorage.ExtStorageClientThread = ExtStorageClientThread

    def tearDown(self):
        PackableStorage.DB = DB
        super(GenericRelStorageTests, self).tearDown()

    def _make_readonly(self):
        # checkWriteMethods in ReadOnlyStorage assumes that
        # the object has an undo() method, even though that's only
        # required if it's IStorageUndoable, aka history-preserving.
        super(GenericRelStorageTests, self)._make_readonly()
        storage = self._storage
        if not hasattr(storage, 'undo'):
            def undo(*args, **kwargs):
                raise ReadOnlyError
            storage.undo = undo # pylint:disable=attribute-defined-outside-init
        return storage

    def checkCurrentObjectTidsRoot(self):
        # Get the root object in place
        db = self._closing(DB(self._storage))
        conn = self._closing(db.open())

        storage = conn._storage
        cursor = storage._load_connection.cursor
        oid_to_tid = storage._adapter.mover.current_object_tids(cursor, [0])
        self.assertEqual(1, len(oid_to_tid))
        self.assertIn(0, oid_to_tid)

        # Ask for many, many objects that don't exist.
        # Force the implementation to loop if that's what it does internally.
        oid_to_tid = storage._adapter.mover.current_object_tids(cursor, range(0, 3523))
        self.assertEqual(1, len(oid_to_tid))
        self.assertIn(0, oid_to_tid)

        # No matching oids.
        oid_to_tid = storage._adapter.mover.current_object_tids(cursor, range(1, 3523))
        self.assertEqual(0, len(oid_to_tid))

        conn.close()
        db.close()

    def checkLen(self):
        # Override the version from BasicStorage because we
        # actually do guarantee to keep track of the counts,
        # within certain limits.

        # len(storage) reports the number of objects.
        # check it is zero when empty
        self.assertEqual(len(self._storage), 0)
        # check it is correct when the storage contains two object.
        # len may also be zero, for storages that do not keep track
        # of this number
        self._dostore(data=PersistentMapping())
        self._dostore(data=PersistentMapping())
        self._storage._adapter.stats.large_database_change()
        self.assertEqual(len(self._storage), 2)

    def checkDropAndPrepare(self):
        # Under PyPy, this test either takes a very long time (PyMySQL)
        # or hangs (psycopg2cffi) longer than I want to wait (10+ minutes).
        # This suggests there's a lock on a particular table (the eighth table we drop)
        # which in turn suggests that there are connections still open and leaked!
        # Running a manual GC seems to fix it. It's hard to reproduce manually because
        # it seems to depend on a particular set of tests being run.
        import gc
        gc.collect()
        gc.collect()

        self._storage._adapter.schema.drop_all()
        self._storage._adapter.schema.prepare()

    def checkCrossConnectionInvalidation(self):
        # Verify connections see updated state at txn boundaries
        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['myobj'] = 'yes'
            c2 = db.open()
            r2 = c2.root()
            self.assertNotIn('myobj', r2)

            storage = c1._storage
            t = transaction.Transaction()
            t.description = u'invalidation test'
            c1.tpc_begin(t)
            c1.commit(t)
            storage.tpc_vote(storage._transaction)
            storage.tpc_finish(storage._transaction)

            self.assertNotIn('myobj', r2)
            c2.sync()
            self.assertIn('myobj', r2)
            self.assertEqual(r2['myobj'], 'yes')
        finally:
            db.close()

    def checkCrossConnectionIsolation(self):
        # Verify MVCC isolates connections
        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['alpha'] = PersistentMapping()
            r1['gamma'] = PersistentMapping()
            transaction.commit()

            # Open a second connection but don't load root['alpha'] yet
            c2 = db.open()
            r2 = c2.root()

            r1['alpha']['beta'] = 'yes'

            storage = c1._storage
            t = transaction.Transaction()
            t.description = u'isolation test 1'
            c1.tpc_begin(t)
            c1.commit(t)
            storage.tpc_vote(storage._transaction)
            storage.tpc_finish(storage._transaction)

            # The second connection will now load root['alpha'], but due to
            # MVCC, it should continue to see the old state.
            self.assertIsNone(r2['alpha']._p_changed)  # A ghost
            self.assertFalse(r2['alpha'])
            self.assertEqual(r2['alpha']._p_changed, 0)

            # make root['alpha'] visible to the second connection
            c2.sync()

            # Now it should be in sync
            self.assertIsNone(r2['alpha']._p_changed)  # A ghost
            self.assertTrue(r2['alpha'])
            self.assertEqual(r2['alpha']._p_changed, 0)
            self.assertEqual(r2['alpha']['beta'], 'yes')

            # Repeat the test with root['gamma']
            r1['gamma']['delta'] = 'yes'

            storage = c1._storage
            t = transaction.Transaction()
            t.description = u'isolation test 2'
            c1.tpc_begin(t)
            c1.commit(t)
            storage.tpc_vote(storage._transaction)
            storage.tpc_finish(storage._transaction)

            # The second connection will now load root[3], but due to MVCC,
            # it should continue to see the old state.
            self.assertIsNone(r2['gamma']._p_changed)  # A ghost
            self.assertFalse(r2['gamma'])
            self.assertEqual(r2['gamma']._p_changed, 0)

            # make root[3] visible to the second connection
            c2.sync()

            # Now it should be in sync
            self.assertIsNone(r2['gamma']._p_changed)  # A ghost
            self.assertTrue(r2['gamma'])
            self.assertEqual(r2['gamma']._p_changed, 0)
            self.assertEqual(r2['gamma']['delta'], 'yes')
        finally:
            db.close()

    def checkResolveConflictBetweenConnections(self, clear_cache=False):
        # Verify that conflict resolution works between storage instances
        # bound to connections.
        obj = ConflictResolution.PCounter()
        obj.inc()
        # Establish a polling state; dostoreNP won't.
        self._storage.poll_invalidations()
        oid = self._storage.new_oid()

        revid1 = self._dostoreNP(oid, data=zodb_pickle(obj))
        self._storage.poll_invalidations()
        # These will both poll and get the state for (oid, revid1)
        # cached at that location, where it will be found during conflict
        # resolution.
        storage1 = self._storage.new_instance()
        storage1.load(oid, '')

        storage2 = self._storage.new_instance()
        storage2.load(oid, '')
        # Remember that the cache stats are shared between instances.
        # The first had to fetch it, the second can use it.
        __traceback_info__ = storage1._cache.stats()
        self.assertEqual(storage1._cache.stats()['hits'], 1)
        storage1._cache.reset_stats()
        if clear_cache:
            storage1._cache.clear(load_persistent=False)
        self.assertEqual(storage1._cache.stats()['hits'], 0)

        obj.inc()
        obj.inc()
        # The effect of committing two transactions with the same
        # pickle is to commit two different transactions relative to
        # revid1 that add two to _value.
        root_storage = self._storage
        try:

            def noConflict(*_args, **_kwargs):
                self.fail("Should be no conflict.")
            storage1.tryToResolveConflict = noConflict
            self._storage = storage1
            _revid2 = self._dostoreNP(oid, revid=revid1, data=zodb_pickle(obj))

            # This one had no conflicts and did no cache work
            self.assertEqual(storage1._cache.stats()['hits'], 0)
            self.assertEqual(storage1._cache.stats()['misses'], 0)

            # This will conflict; we will prefetch everything through the cache,
            # or database, and not the storage's loadSerial.
            def noLoadSerial(*_args, **_kwargs):
                self.fail("loadSerial on the storage should never be called")
            storage2.loadSerial = noLoadSerial
            self._storage = storage2
            _revid3 = self._dostoreNP(oid, revid=revid1, data=zodb_pickle(obj))

            # We don't actually update cache stats at all, however,
            # despite the prefetching.
            cache_stats = storage1._cache.stats()
            __traceback_info__ = cache_stats, clear_cache

            self.assertEqual(cache_stats['misses'], 0)
            self.assertEqual(cache_stats['hits'], 0)

            data, _serialno = self._storage.load(oid, '')
            inst = zodb_unpickle(data)
            self.assertEqual(inst._value, 5)
        finally:
            storage1.close()
            storage2.close()
            self._storage = root_storage

    def checkResolveConflictBetweenConnectionsNoCache(self):
        # If we clear the cache, we can still loadSerial()
        self.checkResolveConflictBetweenConnections(clear_cache=True)

    def check16KObject(self):
        # Store 16 * 1024 bytes in an object, then retrieve it
        data = b'a 16 byte string' * 1024
        oid = self._storage.new_oid()
        self._dostoreNP(oid, data=data)
        got, _ = self._storage.load(oid, '')
        self.assertIsInstance(got, bytes)
        self.assertEqual(got, data)
        self.assertEqual(len(got), len(data))

    def check16MObject(self):
        # Store 16 * 1024 * 1024 bytes in an object, then retrieve it
        data = b'a 16 byte string' * (1024 * 1024)
        oid = self._storage.new_oid()
        self._dostoreNP(oid, data=data)
        got, _serialno = self._storage.load(oid, '')
        self.assertEqual(len(got), len(data))
        self.assertEqual(got, data)

    def check99X1900Objects(self):
        # Store 99 objects each with 1900 bytes.  This is intended
        # to exercise possible buffer overfilling that the batching
        # code might cause.
        data = b'0123456789012345678' * 100
        t = transaction.Transaction()
        self._storage.tpc_begin(t)
        oids = []
        for _ in range(99):
            oid = self._storage.new_oid()
            self._storage.store(oid, b'\0'*8, data, '', t)
            oids.append(oid)
        self._storage.tpc_vote(t)
        self._storage.tpc_finish(t)
        for oid in oids:
            got, _serialno = self._storage.load(oid, '')
            self.assertEqual(len(got), len(data))
            self.assertEqual(got, data)

    def checkPreventOIDOverlap(self):
        # Store an object with a particular OID, then verify that
        # OID is not reused.
        data = b'mydata'
        oid1 = b'\0' * 7 + b'\x0f'
        self._dostoreNP(oid1, data=data)
        oid2 = self._storage.new_oid()
        oid1_int = bytes8_to_int64(oid1)
        oid2_int = bytes8_to_int64(oid2)
        self.assertGreater(
            oid2_int, oid1_int,
            'old OID %r (%d) should be less than new OID %r (%d)'
            % (oid1, oid1_int, oid2, oid2_int))

    def checkNoDuplicateOIDsManyThreads(self):
        # Many threads in many storages can allocate OIDs with
        # no duplicates or overlaps.
        # https://github.com/zodb/relstorage/issues/283
        from itertools import combinations

        thread_count = 11
        oids_per_segment = 578
        segment_count = 3
        total_expected_oids = oids_per_segment * segment_count
        oids_by_thread = [list() for _ in range(thread_count)]

        def allocate_oids(thread_storage, thread_num):
            conn_pool = thread_storage._store_connection_pool
            store_conn = conn_pool.borrow()
            try:
                allocator = thread_storage._oids
                my_oids = oids_by_thread[thread_num]
                for _ in range(segment_count):
                    my_oids.extend(
                        bytes8_to_int64(thread_storage.new_oid())
                        for _ in range(oids_per_segment)
                    )
                    # Periodically call set_min_oid, like the storage does,
                    # to check for interference.
                    with conn_pool.borrowing() as store_conn:
                        allocator.set_min_oid(store_conn, my_oids[-1])
                        store_conn.commit()
            finally:
                self.assertLessEqual(conn_pool.pooled_connection_count, len(threads))
                thread_storage.release()


        threads = [threading.Thread(target=allocate_oids,
                                    args=(self._storage.new_instance(), i))
                   for i in range(thread_count)]
        for t in threads:
            t.start()

        for t in threads:
            t.join(99)

        # All of them are released, so we should be down to only one instance.
        self.assertEqual(1, self._storage._store_connection_pool.instance_count)
        self.assertLessEqual(self._storage._store_connection_pool.pooled_connection_count, 1)

        # They all have the desired length, and each one has no duplicates.
        self.assertEqual(
            [len(s) for s in oids_by_thread],
            [total_expected_oids for _ in range(thread_count)]
        )
        self.assertEqual(
            [len(s) for s in oids_by_thread],
            [len(set(s)) for s in oids_by_thread]
        )

        # They are all disjoint
        for a, b in combinations(oids_by_thread, 2):
            __traceback_info__ = a, b
            a = set(a)
            b = set(b)
            self.assertTrue(a.isdisjoint(b))

        # They are all monotonically increasing.
        for s in oids_by_thread:
            self.assertEqual(
                s,
                sorted(s)
            )

    def checkUseCache(self):
        # Store an object, cache it, then retrieve it from the cache
        self._storage = self.make_storage(
            cache_servers='x:1 y:2',
            cache_module_name=fakecache.__name__,
            cache_prefix='zzz',
        )

        fakecache.data.clear()
        db = DB(self._storage)
        try:
            c1 = db.open()
            self.assertEqual(
                c1._storage._cache.cache.g.client.servers,
                ['x:1', 'y:2'])
            r1 = c1.root()
            # The root state and checkpoints should now be cached.
            # A commit count *might* be cached depending on the ZODB version.
            # (Checkpoints are stored in the cache for the sake of tests/monitoring,
            # but aren't read.)
            # self.assertIn('zzz:checkpoints', fakecache.data)
            # self.assertIsNotNone(db.storage._cache.polling_state.checkpoints)
            self.assertEqual(sorted(fakecache.data.keys())[-1][:10],
                             'zzz:state:')
            r1['alpha'] = PersistentMapping()
            transaction.commit()
            cp_count = 1
            if self.keep_history:
                item_count = 2
            else:
                # The previous root state was automatically invalidated
                # XXX: We go back and forth on that.
                item_count = 2
            item_count += cp_count
            self.assertEqual(len(fakecache.data), item_count)

            oid = r1['alpha']._p_oid
            c1._storage.load(oid, '')
            # Came out of the cache, nothing new
            self.assertEqual(len(fakecache.data), item_count)

            # make a change
            r1['beta'] = 0
            transaction.commit()
            # Once again, history free automatically invalidated.
            # XXX: Depending on my mood.
            item_count += 1
            self.assertEqual(len(fakecache.data), item_count)

            c1._storage.load(oid, '')

            # try to load an object that doesn't exist
            self.assertRaises(KeyError, c1._storage.load, b'bad.oid.', '')
        finally:
            db.close()

    def checkMultipleStores(self):
        # Verify a connection can commit multiple transactions
        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['alpha'] = 1
            transaction.commit()
            r1['alpha'] = 2
            transaction.commit()
        finally:
            db.close()

    def checkLongTransactionDescription(self):
        # Don't trip over long transaction descriptions
        db = DB(self._storage)
        try:
            c = db.open()
            r = c.root()
            r['key'] = 1
            transaction.get().note(u'A long description. ' * 1000)
            transaction.commit()
        finally:
            db.close()

    def checkAutoReconnect(self):
        # Verify auto-reconnect
        db = self._closing(DB(self._storage))

        c1 = db.open()
        r = c1.root()
        r['alpha'] = 1
        transaction.commit()
        c1.close()

        # Going behind its back.
        c1._storage._load_connection.connection.close()
        c1._storage._store_connection_pool.hard_close_all_connections()
        store_pool = c1._storage._store_connection_pool
        self.assertEqual(store_pool.instance_count, 2)
        self.assertLessEqual(store_pool.pooled_connection_count, 1)
        # ZODB5 implicitly calls sync
        # immediately when a connection is opened;
        # fake that here for older releases.
        c2 = db.open()
        self.assertIs(c2, c1)
        c2.sync()
        r = c2.root()
        self.assertEqual(r['alpha'], 1)
        r['beta'] = PersistentMapping()
        c2.add(r['beta']) # Calling new_oid outside of TPC
        transaction.commit()
        c2.close()

        del c1
        del c2

    def checkAutoReconnectOnSync(self):
        # Verify auto-reconnect.
        db = self._closing(DB(self._storage))

        c1 = db.open()
        r = c1.root()

        c1._storage._load_connection.connection.close()
        c1._storage.sync(True)
        # ZODB5 calls sync when a connection is opened. Our monkey
        # patch on a Connection makes sure that works in earlier
        # versions, but we don't have that patch on ZODB5. So test
        # the storage directly. NOTE: The load connection must be open.
        # to trigger the actual sync.

        r = c1.root()
        r['alpha'] = 1
        transaction.commit()
        c1.close()

        c1._storage._load_connection.connection.close()

        c1._storage._store_connection_pool.hard_close_all_connections()
        store_pool = c1._storage._store_connection_pool
        self.assertEqual(store_pool.instance_count, 2)
        self.assertLessEqual(store_pool.pooled_connection_count, 1)


        c2 = db.open()
        self.assertIs(c2, c1)
        self.assertEqual(store_pool.instance_count, 2)
        self.assertLessEqual(store_pool.pooled_connection_count, 1)

        r = c2.root()
        self.assertEqual(r['alpha'], 1)
        r['beta'] = PersistentMapping()
        c2.add(r['beta'])
        transaction.commit()
        c2.close()

        del c1
        del c2

    def checkCachePolling(self):
        storage2 = self.make_storage(zap=False)

        db = DB(self._storage)
        db2 = DB(storage2)
        try:
            # Set up the database.
            tm1 = transaction.TransactionManager()
            c1 = db.open(transaction_manager=tm1)
            r1 = c1.root()
            r1['obj'] = obj1 = PersistentMapping({'change': 0})
            tm1.commit()

            # Load and change the object in an independent connection.
            tm2 = transaction.TransactionManager()
            c2 = db2.open(transaction_manager=tm2)
            r2 = c2.root()
            r2['obj']['change'] = 1
            tm2.commit()
            # Now c2 has delta_after0.
            # self.assertEqual(len(c2._storage._cache.delta_after0), 2)
            c2.close()

            # Change the object in the original connection.
            c1.sync()
            obj1['change'] = 2
            tm1.commit()

            # Close the database connection to c2.
            c2._storage._load_connection.drop()
            self.assertFalse(c2._storage._load_connection)
            # Make the database connection to c2 reopen without polling.
            c2._storage.load(b'\0' * 8, '')
            self.assertTrue(c2._storage._load_connection)

            # Open a connection, which should be the same connection
            # as c2.
            c3 = db2.open(transaction_manager=tm2)
            self.assertTrue(c3 is c2)
            # self.assertEqual(len(c2._storage._cache.delta_after0), 2)

            # Clear the caches (but not delta_after*)
            c3._resetCache()
            c3._storage._cache.cache.flush_all()

            obj3 = c3.root()['obj']
            # Should have loaded the new object.
            self.assertEqual(obj3['change'], 2)

        finally:
            db.close()
            db2.close()

    def checkDoubleCommitter(self):
        # Verify we can store an object that gets committed twice in
        # a single transaction.
        db = DB(self._storage)
        try:
            conn = db.open()
            try:
                conn.root()['dc'] = DoubleCommitter()
                transaction.commit()
                conn2 = db.open()
                self.assertEqual(conn2.root()['dc'].new_attribute, 1)
                conn2.close()
            finally:
                transaction.abort()
                conn.close()
        finally:
            db.close()

    def checkHistoryWithExtension(self):
        # Verify the history method works with transactions that have
        # extended info.
        db = DB(self._storage)
        try:
            conn = db.open()
            try:
                conn.root()['pi'] = 3.14
                transaction.get().setExtendedInfo("digits", 3)
                transaction.commit()
                history = self._storage.history(conn.root()._p_oid)
                self.assertEqual(len(history), 1)
                if self.keep_history:
                    self.assertEqual(history[0]['digits'], 3)
            finally:
                conn.close()
        finally:
            db.close()

    def checkPackBatchLockNoWait(self):
        # Holding the commit lock doesn't interfere with packing.
        #
        # TODO: But what about row locking? Let's add a test
        # that begins a commit and locks some rows and then packs.
        self._storage = self.make_storage(pack_batch_timeout=0)

        adapter = self._storage._adapter
        test_conn, test_cursor = adapter.connmanager.open_for_store()

        db = self._closing(DB(self._storage))
        try:
            # add some data to be packed
            c = self._closing(db.open())
            r = c.root()
            r['alpha'] = PersistentMapping()
            transaction.commit()
            del r['alpha']
            transaction.commit()

            # Pack, with a commit lock held
            now = packtime = time.time()
            while packtime <= now:
                packtime = time.time()
            adapter.locker.hold_commit_lock(test_cursor)
            self._storage.pack(packtime, referencesf)
            adapter.locker.release_commit_lock(test_cursor)
        finally:
            db.close()
            adapter.connmanager.close(test_conn, test_cursor)

    def checkPackKeepNewObjects(self):
        # Packing should not remove objects created or modified after
        # the pack time, even if they are unreferenced.
        db = DB(self._storage)
        try:
            # add some data to be packed
            c = db.open()
            extra1 = PersistentMapping()
            c.add(extra1)
            extra2 = PersistentMapping()
            c.add(extra2)
            transaction.commit()

            # Choose the pack time to be that last committed transaction.
            packtime = c._storage.lastTransactionInt()

            extra2.foo = 'bar'
            extra3 = PersistentMapping()
            c.add(extra3)
            transaction.commit()

            self.assertGreater(c._storage.lastTransactionInt(), packtime)

            self._storage.pack(packtime, referencesf)

            # extra1 should have been garbage collected
            self.assertRaises(KeyError,
                              self._storage.load, extra1._p_oid, '')
            # extra2 and extra3 should both still exist
            self._storage.load(extra2._p_oid, '')
            self._storage.load(extra3._p_oid, '')
        finally:
            db.close()

    def checkPackBrokenPickle(self):
        # Verify the pack stops with the right exception if it encounters
        # a broken pickle.
        # Under Python 2, with zodbpickle, there may be a difference depending
        # on whether the accelerated implementation is in use. Also ,the pure-python
        # version on PyPy can raise IndexError
        from zodbpickle.pickle import UnpicklingError as pUnpickErr
        unpick_errs = (pUnpickErr, IndexError)
        try:
            from zodbpickle.fastpickle import UnpicklingError as fUnpickErr
        except ImportError:
            pass
        else:
            unpick_errs += (fUnpickErr,)


        self._dostoreNP(self._storage.new_oid(), data=b'brokenpickle')
        self.assertRaises(unpick_errs, self._storage.pack,
                          time.time() + 10000, referencesf)

    def checkBackwardTimeTravelWithoutRevertWhenStale(self):
        # If revert_when_stale is false (the default), when the database
        # connection is stale (such as through failover to an
        # asynchronous slave that is not fully up to date), the poller
        # should notice that backward time travel has occurred and
        # raise a ReadConflictError.
        self._storage = self.make_storage(revert_when_stale=False)
        db = DB(self._storage)
        try:
            c = db.open()
            c._storage._adapter.poller.transactions_may_go_backwards = True
            r = c.root()
            r['alpha'] = PersistentMapping()
            transaction.commit()

            # To simulate failover to an out of date async slave, take
            # a snapshot of the database at this point, change some
            # object, then restore the database to its earlier state.

            d = tempfile.mkdtemp()
            try:
                # Snapshot the database.
                fs = FileStorage(os.path.join(d, 'Data.fs'))
                fs.copyTransactionsFrom(c._storage)

                # Change data in it.
                r['beta'] = PersistentMapping()
                transaction.commit()
                self.assertTrue('beta' in r)

                # Revert the data.
                # We must use a separate, unrelated storage object to do this,
                # because our storage object is smart enough to notice that the data
                # has been zapped and revert caches for all connections and
                # ZODB objects when we invoke this API.
                storage_2 = self.make_storage(zap=False)
                storage_2.zap_all(reset_oid=False, slow=True)
                storage_2.copyTransactionsFrom(fs)
                storage_2.close()
                del storage_2
                fs.close()
                del fs
            finally:
                shutil.rmtree(d)

            # Sync, which will call poll_invalidations().
            c.sync()

            # Try to load an object, which should cause ReadConflictError.
            r._p_deactivate()
            with self.assertRaises(ReadConflictError):
                r.__getitem__('beta')

        finally:
            db.close()

    def checkBackwardTimeTravelWithRevertWhenStale(self):
        # If revert_when_stale is true, when the database
        # connection is stale (such as through failover to an
        # asynchronous slave that is not fully up to date), the poller
        # should notice that backward time travel has occurred and
        # invalidate all objects that have changed in the interval.
        self._storage = self.make_storage(revert_when_stale=True)

        db = DB(self._storage)
        try:
            transaction.begin()
            c = db.open()
            r = c.root()
            r['alpha'] = PersistentMapping()
            transaction.commit()

            # To simulate failover to an out of date async slave, take
            # a snapshot of the database at this point, change some
            # object, then restore the database to its earlier state.

            d = tempfile.mkdtemp()
            try:
                transaction.begin()
                fs = FileStorage(os.path.join(d, 'Data.fs'))
                fs.copyTransactionsFrom(c._storage)

                r['beta'] = PersistentMapping()
                transaction.commit()
                self.assertTrue('beta' in r)

                c._storage.zap_all(reset_oid=False, slow=True)
                c._storage.copyTransactionsFrom(fs)

                fs.close()
            finally:
                shutil.rmtree(d)

            # r should still be in the cache.
            self.assertTrue('beta' in r)

            # Now sync, which will call poll_invalidations().
            c.sync()

            # r should have been invalidated
            self.assertEqual(r._p_changed, None)

            # r should be reverted to its earlier state.
            self.assertFalse('beta' in r)

        finally:
            db.close()

    @util.skipOnAppveyor("Random failures")
    # https://ci.appveyor.com/project/jamadden/relstorage/build/1.0.75/job/32uu4xdp5mubqma8
    def checkBTreesLengthStress(self):
        # BTrees.Length objects are unusual Persistent objects: they
        # have a conflict resolution algorithm that cannot fail, so if
        # we do get a failure it's due to a problem with us.
        # Unfortunately, tryResolveConflict hides all underlying exceptions
        # so we have to enable logging to see them.
        from relstorage.adapters.interfaces import UnableToAcquireLockError
        from ZODB.ConflictResolution import logger as CRLogger
        from BTrees.Length import Length
        import BTrees
        from six import reraise

        def log_err(*args, **kwargs): # pylint:disable=unused-argument
            import sys
            reraise(*sys.exc_info())

        CRLogger.debug = log_err
        CRLogger.exception = log_err

        updates_per_thread = 50
        thread_count = 4
        lock_errors = []

        self.maxDiff = None
        db = DB(self._storage)
        try:
            c = db.open()
            try:
                root = c.root()
                root['length'] = Length()
                # XXX: Eww! On MySQL, if we try to take a shared lock on
                # OID 0, and a write lock on OID 1, we fail with a deadlock
                # error. It seems that taking the shared lock on 0 also takes a shared
                # lock on 1 --- somehow. Because they're adjacent to each other?
                # I don't know. We have to add some space between them to be sure
                # that doesn't happen. On MySQL 5.7, just 10 extra items was enough.
                # On MySQL 8, we had to add more.
                for i in range(50):
                    root[i] = BTrees.OOBTree.BTree() # pylint:disable=no-member
                transaction.commit()
            except:
                transaction.abort()
                raise
            finally:
                c.close()

            def updater():
                for _ in range(updates_per_thread):
                    thread_c = db.open()
                    __traceback_info__ = thread_c._storage
                    try:
                        # Perform readCurrent on an object not being modified.
                        # This adds stress to databases that use separate types of locking
                        # for modified and current objects. It was used to discover
                        # bugs in gevent+MySQL and plain MySQLdb against both 5.7 and 8.
                        root = thread_c.root()
                        root._p_activate() # unghost; only non-ghosts can readCurrent
                        root._p_jar.readCurrent(root)
                        root['length'].change(1)
                        time.sleep(random.random() * 0.05)
                        try:
                            transaction.commit()
                        except UnableToAcquireLockError as e:
                            lock_errors.append((type(e), str(e)))
                            transaction.abort()
                            raise
                    finally:
                        thread_c.close()

            threads = []
            for _ in range(thread_count):
                t = threading.Thread(target=updater)
                threads.append(t)
            for t in threads:
                t.start()
            for t in threads:
                t.join(120)

            self.assertEqual(lock_errors, [])

            c = db.open()
            try:
                self.assertEqual(c.root()['length'](),
                                 updates_per_thread * thread_count)
            finally:
                transaction.abort()
                c.close()

        finally:
            db.close()
            del CRLogger.debug
            del CRLogger.exception


    def checkAfterCompletion(self):
        # The after completion method, which can only be called
        # outside of 2-phase commit is otherise equivalent to calling
        # tpc_abort.
        from ZODB.interfaces import IMVCCAfterCompletionStorage
        self._storage = self.make_storage(revert_when_stale=False)

        with mock.patch.object(self._storage._load_connection,
                               'rollback_quietly') as rb:
            self._storage.afterCompletion()
            rb.assert_called_with()

        self.assertTrue(
            IMVCCAfterCompletionStorage.providedBy(self._storage))

    def checkConfigureViaZConfig(self):
        replica_fn = None
        replica_conf = ''
        if util.DEFAULT_DATABASE_SERVER_HOST == util.STANDARD_DATABASE_SERVER_HOST:
            replica_fn = self.get_adapter_zconfig_replica_conf()
            if replica_fn:
                replica_conf = 'replica-conf ' + self.get_adapter_zconfig_replica_conf()

        conf = u"""
        %import relstorage
        <zodb main>
            <relstorage>
            name xyz
            read-only false
            keep-history {KEEP_HISTORY}
            {REPLICA_CONF}
            blob-dir .
            blob-cache-size-check-external true
            blob-cache-size 100MB
            blob-chunk-size 10MB
            cache-local-dir-read-count 12
            cache-local-dir-write-max-size 10MB
            {ADAPTER}
            </relstorage>
        </zodb>
        """.format(
            KEEP_HISTORY='true' if self.keep_history else 'false',
            REPLICA_CONF=replica_conf,
            ADAPTER=self.get_adapter_zconfig()
        )

        __traceback_info__ = conf

        schema_xml = u"""
        <schema>
        <import package="ZODB"/>
        <section type="ZODB.database" name="main" attribute="database"/>
        </schema>
        """
        import ZConfig
        from io import StringIO
        from ZODB.interfaces import IBlobStorageRestoreable
        from relstorage.adapters.interfaces import IRelStorageAdapter
        from relstorage.blobhelper.interfaces import ICachedBlobHelper
        from hamcrest import assert_that
        from nti.testing.matchers import validly_provides
        schema = ZConfig.loadSchemaFile(StringIO(schema_xml))
        config, _ = ZConfig.loadConfigFile(schema, StringIO(conf))

        db = config.database.open()
        try:
            storage = db.storage
            assert_that(storage, validly_provides(IBlobStorageRestoreable))
            self.assertEqual(storage.isReadOnly(), False)
            self.assertEqual(storage.getName(), "xyz")
            assert_that(storage.blobhelper, validly_provides(ICachedBlobHelper))
            self.assertIn('_External', str(storage.blobhelper.cache_checker))

            adapter = storage._adapter
            self.assertIsInstance(adapter, self.get_adapter_class())
            assert_that(adapter, validly_provides(IRelStorageAdapter))
            self.verify_adapter_from_zconfig(adapter)
            self.assertEqual(adapter.keep_history, self.keep_history)
            if replica_fn:
                self.assertEqual(
                    adapter.connmanager.replica_selector.replica_conf,
                    replica_fn)
            self.assertEqual(storage._options.blob_chunk_size, 10485760)
        finally:
            db.close()

    def checkGeventSwitchesOnOpen(self):
        # We make some queries when we open; if the driver is gevent
        # capable, that should switch.
        driver = self._storage._adapter.driver
        if not driver.gevent_cooperative():
            raise unittest.SkipTest("Driver %s not gevent capable" % (driver,))

        from gevent.util import assert_switches
        with assert_switches():
            self.open()

    #####
    # Prefetch Tests
    #####

    def checkPrefetch(self):
        db = DB(self._storage)
        conn = db.open()

        mapping = conn.root()['key'] = PersistentMapping()

        transaction.commit()

        item_count = 3

        # The new state for the root invalidated the old state,
        # and since there is no other connection that might be using it,
        # we drop it from the cache.
        item_count = 2
        self.assertEqual(item_count, len(self._storage._cache))
        tid = bytes8_to_int64(mapping._p_serial)
        d = self._storage._cache.local_client._cache
        self.assertEqual(d[0].max_tid, tid)
        self.assertEqual(d[1].max_tid, tid)
        self._storage._cache.clear()

        self.assertEmpty(self._storage._cache)

        conn.prefetch(z64, mapping)

        self.assertEqual(2, len(self._storage._cache))

        # second time is a no-op
        conn.prefetch(z64, mapping)
        self.assertEqual(2, len(self._storage._cache))

    ######
    # Parallel Commit Tests
    ######

    @skipIfNoConcurrentWriters
    def checkCanVoteAndCommitWhileOtherStorageVotes(self):
        storage1 = self._closing(self._storage.new_instance())
        storage2 = self._closing(self._storage.new_instance())

        # Bring them both into tpc_vote phase. Before parallel commit,
        # this would have blocked as the first storage took the commit lock
        # in tpc_vote.
        txs = {}
        for storage in (storage1, storage2):
            data = zodb_pickle(MinPO(str(storage)))
            t = TransactionMetaData()
            txs[storage] = t

            storage.tpc_begin(t)
            oid = storage.new_oid()
            storage.store(oid, None, data, '', t)
            storage.tpc_vote(t)

        # The order we choose to finish is the order of the returned
        # tids.
        tid1 = storage2.tpc_finish(txs[storage2])
        tid2 = storage1.tpc_finish(txs[storage1])

        self.assertGreater(tid2, tid1)

        storage1.close()
        storage2.close()

    def checkCanLoadObjectStateWhileBeingModified(self):
        # Get us an object in the database
        storage1 = self._closing(self._storage.new_instance())
        data = zodb_pickle(MinPO(str(storage1)))
        t = TransactionMetaData()
        storage1.tpc_begin(t)
        oid = storage1.new_oid()

        storage1.store(oid, None, data, '', t)
        storage1.tpc_vote(t)
        initial_tid = storage1.tpc_finish(t)

        storage1.release()
        del storage1

        self._storage._cache.clear(load_persistent=False)

        storage1 = self._closing(self._storage.new_instance())

        # Get a completely independent storage, not sharing a cache
        storage2 = self._closing(self.make_storage(zap=False))

        # First storage attempts to modify the oid.
        t = TransactionMetaData()
        storage1.tpc_begin(t)
        storage1.store(oid, initial_tid, data, '', t)
        # And locks the row.
        storage1.tpc_vote(t)

        # storage2 would like to read the old row.
        loaded_data, loaded_tid = storage2.load(oid)

        self.assertEqual(loaded_data, data)
        self.assertEqual(loaded_tid, initial_tid)

        # Commit can now happen.
        tid2 = storage1.tpc_finish(t)
        self.assertGreater(tid2, initial_tid)

        storage1.close()
        storage2.close()

    ###
    # IStorageCurrentRecordIteration tests
    ###

    def check_record_iternext_basic(self, start_oid_int=None):
        # Based on code from FileStorage tests

        db = DB(self._storage)
        conn = db.open()
        conn.root()['abc'] = MinPO('abc')
        conn.root()['xyz'] = MinPO('xyz')
        transaction.commit()

        # Now, add some additional revisions. This proves that we iterate latest reconds,
        # not all transactions.
        conn.root()['abc'].value = 'def'
        conn.root()['xyz'].value = 'ghi'
        transaction.commit()
        conn.close()

        storage2 = self._closing(self._storage.new_instance())

        # The special case: convert to byte OID
        token = None if start_oid_int is None else int64_to_8bytes(start_oid_int)

        # (0, 1, 2) by default, or, e.g, (1, 2)
        expected_oids = range(start_oid_int or 0, 3)
        if not expected_oids:
            assert start_oid_int > 3
            # Call at least once.
            expected_oids = (0,)
        record_count = 0
        for x in expected_oids:
            oid, tid, data, next_token = self._storage.record_iternext(token)
            record_count += 1
            self.assertEqual(oid, int64_to_8bytes(x))
            token = next_token
            expected_data, expected_tid = storage2.load(oid)
            self.assertEqual(expected_data, data)
            self.assertEqual(expected_tid, tid)
            if x == 2:
                check_token = self.assertIsNone
            else:
                check_token = self.assertIsNotNone
            check_token(token)
        self.assertEqual(len(expected_oids), record_count)

    def check_record_iternext_token_0(self):
        # Passing a starting token.
        self.check_record_iternext_basic(0)

    def check_record_iternext_token_1(self):
        # Gets a subset.
        self.check_record_iternext_basic(1)

    def check_record_iternext_too_large_oid(self):
        with self.assertRaises(StopIteration):
            self.check_record_iternext_basic(10)


class AbstractRSZodbConvertTests(StorageCreatingMixin,
                                 FSZODBConvertTests,
                                 # This one isn't cooperative in
                                 # setUp(), so it needs to be last.
                                 ZODB.tests.util.TestCase):
    keep_history = True
    filestorage_name = 'source'
    relstorage_name = 'destination'
    filestorage_file = None

    def setUp(self):
        super(AbstractRSZodbConvertTests, self).setUp()
        cfg = """
        %%import relstorage
        %%import zc.zlibstorage
        <zlibstorage %s>
        <filestorage>
            path %s
        </filestorage>
        </zlibstorage>
        <zlibstorage %s>
        <relstorage>
            %s
            cache-prefix %s
            cache-local-dir %s
        </relstorage>
        </zlibstorage>
        """ % (
            self.filestorage_name,
            self.filestorage_file,
            self.relstorage_name,
            self.get_adapter_zconfig(),
            self.relstorage_name,
            os.path.abspath('.'),
        )
        self._write_cfg(cfg)

        self.make_storage(zap=True).close()

    def _wrap_storage(self, storage):
        return self._closing(ZlibStorage(storage))

    def _create_dest_storage(self):
        return self._wrap_storage(super(AbstractRSZodbConvertTests, self)._create_dest_storage())

    def _create_src_storage(self):
        return self._wrap_storage(super(AbstractRSZodbConvertTests, self)._create_src_storage())

    def test_new_instance_still_zlib(self):
        storage = self._closing(self.make_storage())
        new_storage = self._closing(storage.new_instance())
        self.assertIsInstance(new_storage,
                              ZlibStorage)

        self.assertIn('_crs_untransform_record_data', storage.base.__dict__)
        self.assertIn('_crs_transform_record_data', storage.base.__dict__)

        self.assertIn('_crs_untransform_record_data', new_storage.base.__dict__)
        self.assertIn('_crs_transform_record_data', new_storage.base.__dict__)

class AbstractRSDestZodbConvertTests(AbstractRSZodbConvertTests):

    zap_supported_by_dest = True

    @property
    def filestorage_file(self):
        return self.srcfile

    def _create_dest_storage(self):
        return self._closing(self.make_storage(cache_prefix=self.relstorage_name, zap=False))

class AbstractRSSrcZodbConvertTests(AbstractRSZodbConvertTests):

    filestorage_name = 'destination'
    relstorage_name = 'source'

    @property
    def filestorage_file(self):
        return self.destfile

    def _create_src_storage(self):
        return self._closing(self.make_storage(cache_prefix=self.relstorage_name, zap=False))

class AbstractIDBOptionsTest(unittest.TestCase):

    db_options = None

    def test_db_options_compliance(self):
        from hamcrest import assert_that
        from nti.testing.matchers import validly_provides

        from relstorage.adapters.interfaces import IDBDriverOptions
        from relstorage.adapters.interfaces import IDBDriverFactory
        __traceback_info__ = self.db_options
        assert_that(self.db_options, validly_provides(IDBDriverOptions))

        for factory in self.db_options.known_driver_factories():
            assert_that(factory, validly_provides(IDBDriverFactory))

class AbstractIDBDriverTest(unittest.TestCase):

    driver = None

    def test_db_driver_compliance(self):
        from hamcrest import assert_that
        from nti.testing.matchers import validly_provides

        from relstorage.adapters.interfaces import IDBDriver
        __traceback_info__ = self.driver
        assert_that(self.driver, validly_provides(IDBDriver))


class DoubleCommitter(Persistent):
    """A crazy persistent class that changes self in __getstate__"""
    def __getstate__(self):
        if not hasattr(self, 'new_attribute'):
            self.new_attribute = 1 # pylint:disable=attribute-defined-outside-init
        return Persistent.__getstate__(self)


def _close_and_clean_storage(storage):
    try:
        storage.close()
        storage.cleanup()
    except Exception: # pylint:disable=broad-except
        pass


class AbstractToFileStorage(RelStorageTestBase):
    # Subclass this and set:
    # - keep_history = True; and
    # - A base class of UndoableRecoveryStorage
    #
    # or
    # - keep_history = False; and
    # A base class of BasicRecoveryStorage

    # We rely on being placed in a temporary directory by a super
    # class that will be cleaned up by tearDown().

    def setUp(self):
        super(AbstractToFileStorage, self).setUp()
        # Use the abspath so that even if we close it after
        # we've returned to our original directory (e.g.,
        # close is run as part of addCleanup(), which happens after
        # tearDown) we don't write index files into the original directory.
        self._dst_path = os.path.abspath(self.rs_temp_prefix + 'Dest.fs')
        self.__dst = None

    @property
    def _dst(self):
        if self.__dst is None:
            self.__dst = FileStorage(self._dst_path, create=True)
            # On Windows, though, this could be too late: We can't remove
            # files that are still open, and zope.testing.setupstack
            # was asked to remove the temp dir as part of tearing itself down;
            # cleanups run after tearDown runs (which is when the setupstack runs.)
            self.addCleanup(_close_and_clean_storage, self.__dst)
        return self.__dst

    def tearDown(self):
        if hasattr(self.__dst, 'close'):
            _close_and_clean_storage(self.__dst)
        self.__dst = 42 # Not none so we don't try to create.
        super(AbstractToFileStorage, self).tearDown()

    def new_dest(self):
        return self._closing(FileStorage(self._dst_path))


class AbstractFromFileStorage(RelStorageTestBase):
    # As for AbstractToFileStorage

    def setUp(self):
        super(AbstractFromFileStorage, self).setUp()
        self._src_path = os.path.abspath(self.rs_temp_prefix + 'Source.fs')
        self.__dst = None

    def make_storage_to_cache(self):
        return FileStorage(self._src_path, create=True)

    @property
    def _dst(self):
        if self.__dst is None:
            self.__dst = self.make_storage()
            self.addCleanup(_close_and_clean_storage, self.__dst)
        return self.__dst

    def tearDown(self):
        if hasattr(self.__dst, 'close'):
            _close_and_clean_storage(self.__dst)

        self.__dst = 42 # Not none so we don't try to create.
        super(AbstractFromFileStorage, self).tearDown()

    def new_dest(self):
        return self._dst



##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Helpers for drivers
"""

from __future__ import print_function

import importlib
import sys
import os

from zope.interface import directlyProvides
from zope.interface import implementer

from .._compat import PYPY
from .._compat import PY3
from .._compat import casefold
from .._util import positive_integer
from .._util import consume

from .interfaces import IDBDriver
from .interfaces import IDBDriverFactory
from .interfaces import IDBDriverOptions
from .interfaces import DriverNotAvailableError
from .interfaces import NoDriversAvailableError
from .interfaces import ReplicaClosedException
from .interfaces import UnknownDriverError

logger = __import__('logging').getLogger(__name__)

def _select_driver(options, driver_options):
    driver = _select_driver_by_name(options.driver, driver_options)
    driver.configure_from_options(options)
    return driver

def _select_driver_by_name(driver_name, driver_options):
    driver_name = driver_name or 'auto'
    driver_name = casefold(driver_name)
    accept_any_driver = driver_name == 'auto'
    # XXX: For testing, we'd like to be able to prohibit the use of auto.
    ex_strs = {} # {driver_name: ex}
    for factory in driver_options.known_driver_factories():
        exact_match = casefold(factory.driver_name) == driver_name
        if accept_any_driver or exact_match:
            try:
                result = factory()
            except DriverNotAvailableError as e:
                if not accept_any_driver:
                    e.driver_options = driver_options
                    raise
                ex_strs[factory.driver_name] = str(e)
            else:
                logger.debug("Using driver %s for requested name %r", result, driver_name)
                return result

    # Well snap, no driver. Either we would take any driver,
    # and none were available, or we needed an exact driver that
    # wasn't found
    error = NoDriversAvailableError if accept_any_driver else UnknownDriverError
    raise error(driver_name, driver_options, ex_strs)

class DriverNotImportableError(DriverNotAvailableError,
                               ImportError):
    "When the module can't be imported."

class AbstractModuleDriver(object):
    """
    Base implementation of a driver, based on a module, as used in DBAPI.

    Subclasses must provide:

    - ``MODULE_NAME`` property.
    - ``__name__`` property
    - Implementation of ``get_driver_module``; this should import the
      module at runtime.
    """

    #: The name of the DB-API module to import.
    MODULE_NAME = None

    #: The name written in config files
    __name__ = None

    #: Can this module be used on PyPy?
    AVAILABLE_ON_PYPY = True

    #: Set this to false if your subclass can do static checks
    #: at import time to determine it should not be used.
    #: Helpful for things like Python version detection.
    STATIC_AVAILABLE = True

    #: Priority of this driver, when available. Lower is better.
    #: (That is, first choice should have value 1, and second choice value
    #: 2, and so on.)
    PRIORITY = 100

    #: Priority of this driver when running on PyPy. Lower is better.
    PRIORITY_PYPY = 100

    #: Class attribute. If set to a true value (not the default),
    #: ask the underlying driver to work in as strict a mode as possible
    #: when it comes to detecting programming errors.
    #:
    #: Typically set by tests. Most drivers do not have a stricter mode
    #: that can be enabled.
    STRICT = False

    # Can this driver work with gevent?
    _GEVENT_CAPABLE = False
    # Does this driver need the socket module patched?
    # Only checked if _GEVENT_CAPABLE is set to True.
    _GEVENT_NEEDS_SOCKET_PATCH = True

    #: The size we request cursor's from our :meth:`cursor` method
    #: to fetch from ``fetchmany`` and (hopefully) iteration (which is a
    #: DB-API extension. We default to 1024, but the environment variable
    #: RS_CURSOR_ARRAYSIZE can be set to an int to change this default.
    #: Individual drivers *might* choose a different default.
    cursor_arraysize = positive_integer(
        os.environ.get('RS_CURSOR_ARRAYSIZE', '1024')
    )

    DriverNotAvailableError = DriverNotAvailableError

    # Can the driver support the full range of a 64-bit unsigned ID for
    # OID and TID parameters?
    supports_64bit_unsigned_id = True

    def __init__(self):
        if PYPY and not self.AVAILABLE_ON_PYPY:
            raise self.DriverNotAvailableError(self.__name__)
        if not self.STATIC_AVAILABLE:
            raise self.DriverNotAvailableError(self.__name__)
        try:
            self.driver_module = mod = self.get_driver_module()
        except ImportError as ex:
            logger.debug(
                "Attempting to load driver named %r from %r failed; if no driver was specified, "
                "or the driver was set to 'auto', there may be more drivers to attempt.",
                self.__name__, self.MODULE_NAME,
                exc_info=True)
            raise DriverNotImportableError(self.__name__, reason=str(ex))


        self.disconnected_exceptions = (mod.OperationalError,
                                        mod.InterfaceError,
                                        ReplicaClosedException)
        self.close_exceptions = self.disconnected_exceptions + (mod.ProgrammingError,)
        self.lock_exceptions = (mod.DatabaseError,)
        # If we try to do something very wrong, a bug in our code,
        # we *should* get a ProgrammingError. Unfortunately, some drivers
        # raise ProgrammingError for other things, such as failing to get a lock.
        self.illegal_operation_exceptions = (mod.ProgrammingError,)
        self.use_replica_exceptions = (mod.OperationalError,)
        self.Binary = mod.Binary
        self._connect = mod.connect
        self.priority = self.PRIORITY if not PYPY else self.PRIORITY_PYPY

    def connect(self, *args, **kwargs):
        return self._connect(*args, **kwargs)

    def get_driver_module(self):
        """Import and return the driver module."""
        return importlib.import_module(self.MODULE_NAME)

    def gevent_cooperative(self):
        # Return whether this driver is cooperative with gevent.
        # This takes into account whether the system is
        # and/or needs to be monkey-patched
        if not self._GEVENT_CAPABLE:
            return False
        if self._GEVENT_NEEDS_SOCKET_PATCH:
            return self._sockets_gevent_monkey_patched()
        return True

    def configure_from_options(self, options): # pylint:disable=unused-argument
        """Default implementation; does nothing."""

    def _sockets_gevent_monkey_patched(self):
        # Return whether the socket module has been monkey-patched
        # by gevent
        try:
            from gevent import monkey
        except ImportError: # pragma: no cover
            return False
        else:
            # some versions of gevent have a bug where if we're monkey-patched
            # on the command line using python -m gevent.monkey /path/to/testrunner ...
            # it doesn't report being monkey-patched.
            import socket
            return monkey.is_module_patched('socket') or 'gevent' in repr(socket.socket)

    # Common compatibility shims, overriden as needed.

    def set_autocommit(self, conn, value):
        conn.autocommit(value)

    def cursor(self, conn, server_side=False): # pylint:disable=unused-argument
        cur = conn.cursor()
        cur.arraysize = self.cursor_arraysize
        return cur

    def debug_connection(self, conn, *extra): # pragma: no cover
        print(conn, *extra)

    def get_messages(self, conn): # pragma: no cover pylint:disable=unused-argument
        return ()

    message_logger = logger

    def __transaction_boundary(self, conn, meth):
        meth()
        messages = self.get_messages(conn)
        for msg in messages:
            self.message_logger.debug("Message from the RDBMS: %s", msg.strip())

    def commit(self, conn, cursor=None): #  pylint:disable=unused-argument
        self.__transaction_boundary(conn, conn.commit)

    def rollback(self, conn):
        self.__transaction_boundary(conn, conn.rollback)

    def connection_may_need_rollback(self, conn): # pylint:disable=unused-argument
        return True

    connection_may_need_commit = connection_may_need_rollback

    def synchronize_cursor_for_rollback(self, cursor):
        """Exceptions here are ignored, we don't know what state the cursor is in."""
        # psycopg2 raises ProgrammingError if we rollback when no results
        # are present on the cursor. mysql-connector-python raises
        # InterfaceError. OTOH, mysqlclient raises nothing and even wants
        # it in certain circumstances.

        if cursor is not None:
            try:
                consume(cursor)
            except Exception: # pylint:disable=broad-except
                pass

    # Things that can be recognized as a pickled state,
    # passed to an io.BytesIO reader, and unpickled.

    # Py MySQL Connector/Python returns a bytearray, whereas
    # C MySQL Connector/Python returns bytes.
    # sqlite uses buffer on Py2 and memoryview on Py3.

    # Keep these ordered with the most common at the front;
    # Python does a linear traversal of type checks.
    state_types = (bytes, bytearray)

    def binary_column_as_state_type(self, data):
        if isinstance(data, self.state_types) or data is None:
            return data
        __traceback_info__ = type(data), data
        raise TypeError("Unknown binary state column")


    def binary_column_as_bytes(self, data):
        # Take the same inputs as `as_state_type`, but turn them into
        # actual bytes. This includes None and empty bytes, which becomes
        # the literal b'';
        # XXX: TODO: We don't need all these checks up here. Just the common ones,
        # move everything else to specific drivers.
        if data is None or not data:
            return b''
        if isinstance(data, bytes):
            return data
        if isinstance(data, memoryview):
            return data.tobytes()
        # Everything left we convert with the bytes() construtor.
        # That would be buffer and bytearray
        __traceback_info__ = data, type(data)
        return bytes(data)

    def enter_critical_phase_until_transaction_end(self, connection, cursor):
        """Default implementation; does nothing."""

    def is_in_critical_phase(self, connection, cursor):
        """Default implementation; returns a false value."""

    def exit_critical_phase(self, connection, cursor):
        "Default implementation; does nothing."

class MemoryViewBlobDriverMixin(object):
    # psycopg2 is smart enough to return memoryview or buffer on
    # Py3/Py2, respectively, for BYTEa columns. sqlite3 does exactly
    # the same for BLOB columns (on Python 2; on Python 3 it returns
    # bytes instead of buffer), and defines ``Binary`` that way as
    # well.

    # memoryview can't be passed to bytes() on Py2 or Py3, but it can
    # be passed to cStringIO.StringIO() or io.BytesIO() ---
    # unfortunately, memoryviews, at least, don't like going to
    # io.BytesIO() on Python 3, and that's how we unpickle states. So
    # while ideally we'd like to keep it that way, to save a copy, we
    # are forced to make the copy. Plus there are tests that like to
    # directly compare bytes.

    if PY3:
        def binary_column_as_state_type(self, data):
            if data:
                # Calling 'bytes()' on a memoryview in Python 3 does
                # nothing useful.
                data = data.tobytes()
            return data
    else:
        def binary_column_as_state_type(self, data):
            if data:
                data = bytes(data)
            return data


@implementer(IDBDriverFactory)
class _ClassDriverFactory(object):

    def __init__(self, driver_type):
        self.driver_type = driver_type
        # Getting the name is tricky, the class wants to shadow it.
        self.driver_name = driver_type.__dict__.get('__name__') or driver_type.__name__

    def check_availability(self):
        try:
            self.driver_type()
        except DriverNotAvailableError:
            return False
        return True

    def __call__(self):
        return self.driver_type()

    def __eq__(self, other):
        return (casefold(self.driver_name), self.driver_type) == (
            casefold(other.driver_name), other.driver_type)

    def __hash__(self):
        return hash((casefold(self.driver_name), self.driver_type))

    def __getattr__(self, name):
        return getattr(self.driver_type, name)


def implement_db_driver_options(name, *driver_modules):
    """
    Helper function to be called at a module scope to
    make it implement ``IDBDriverOptions``.

    :param str name: The value of ``__name__``.
    :param driver_modules: Each of these names a module that has
        one or more implementations of ``IDBDriver`` in it,
        as named in their ``__all__`` attribute.

    """

    module = sys.modules[name]

    driver_factories = set()
    for driver_module in driver_modules:
        driver_module = importlib.import_module('.' + driver_module,
                                                name)
        for factory in driver_module.__all__:
            factory = getattr(driver_module, factory)
            if IDBDriver.implementedBy(factory): # pylint:disable=no-value-for-parameter
                driver_factories.add(_ClassDriverFactory(factory))

    module.known_driver_factories = lambda: sorted(
        driver_factories,
        key=lambda factory: factory.PRIORITY if not PYPY else factory.PRIORITY_PYPY,
    )

    directlyProvides(module, IDBDriverOptions)

    module.select_driver = lambda driver_name=None: _select_driver_by_name(driver_name,
                                                                           sys.modules[name])


class _NoGeventDriverMixin(object):
    import time as gevent

    def get_driver_module(self):
        raise ImportError("Could not import gevent")

class _NoGeventConnectionMixin(object):
    gevent_hub = None
    gevent_read_watcher = None
    gevent_write_watcher = None
    gevent_sleep = None

try:
    import gevent
except ImportError:
    GeventDriverMixin = _NoGeventDriverMixin
    GeventConnectionMixin = _NoGeventConnectionMixin
else:
    from gevent.select import select as gselect
    from gevent import monkey as gmonkey
    from gevent.socket import wait
    get_hub = gevent.get_hub

    blocking_select = gmonkey.get_original('select', 'select')
    del gmonkey

    class GeventDriverMixin(object):
        gevent = gevent

    class GeventConnectionMixin(_NoGeventConnectionMixin):
        """
        Helper for a connection that waits using gevent.

        Subclasses must provide a ``fileno()`` method. The usual
        pattern for executing a query would then be something like
        this::

            query = format_query_to_bytes(...)
            self.gevent_wait_write()
            self.send_query()

            self.gevent_wait_read()
            self.read_results()

        It is important that ``send_query`` do nothing but put bytes
        on the wire. It must not include any attempt to wait for a
        response from the database, especially if that response could
        take an arbitrary amount of time or block. (Of course, if
        ``send_query`` and ``read_results`` can arrange to use gevent
        waiting functions too, you'll have finer control. This example
        is all-or-nothing. Sometimes its easy to handle
        ``read_results`` in a looping function using a server-side
        cursor.)

        The ``gevent_wait_read`` and ``gevent_wait_write`` functions
        are implemented using :func:`gevent.socket.wait`. That
        function always takes a full iteration of the event loop to
        determine whether a file descriptor is ready; it always yields
        control to other greenlets immediately. gevent's own sockets
        don't work that way; instead they try to read/write and catch
        the resulting EAGAIN exception. Only after that do they yield
        to the event loop. This is for good reason: eliminating
        unnecessary switches can lead to higher throughput.

        Here, a pass through the event loop can be risky. If we send a
        request that establishes database locks that will require
        further action from the greenlet to relinquish, those will
        come into being (potentially blocking other greenlets in the
        same or different processes) sometime between when
        ``send_query`` is entered and when ``gevent_wait_read`` exits.
        If, for any reason, a different greenlet runs while we have
        yielded to the event loop and blocks on a resource we own that
        is not gevent cooperative (a non-monkey-patched lock, a
        different database) we'll never regain control. And thus we'll
        never be able to make forward progress and release those
        locks. Since they're shared locks, that could harm arbitrary
        machines in the cluster.

        Thus, by default we perform a similar optimization as gevent sockets: we
        first check to see if the file descriptor is ready and only
        yield to the event loop if it isn't. The cost is an extra
        system call to ``select``. For write requests, we could be
        able to assume that they are always ready (depending on the
        nature of the protocol); if that's so, override
        :meth:`gevent_check_write`. The same goes for
        :meth:`gevent_check_read`. This doesn't eliminate the problem,
        but it should substantially reduce the chances of it
        happening.
        """
        gevent_sleep = staticmethod(gevent.sleep)

        # We don't have a fileno() method ourself,
        # because we're often mixed in at the first.

        def close(self):
            self.__close_watchers()
            super(GeventConnectionMixin, self).close()

        def __check_watchers(self):
            # We can be used from more than one thread in a sequential
            # fashion.
            hub = get_hub()
            if hub is not self.gevent_hub:
                self.__close_watchers()

                # XXX: Any db drivers that silently reconnect? That would change the fileno.
                fileno = self.fileno()
                hub = self.gevent_hub = get_hub()
                self.gevent_read_watcher = hub.loop.io(fileno, 1)
                self.gevent_write_watcher = hub.loop.io(fileno, 2)

        def __close_watchers(self):
            if self.gevent_read_watcher is not None:
                self.gevent_read_watcher.close()
                self.gevent_write_watcher.close()
                self.gevent_hub = None

        def gevent_check_read(self,):
            if gselect([self], (), (), 0)[0]:
                # gselect inserts a call to gevent.sleep() to spin
                # the event loop. If we don't do that, then one connection
                # can monopolize the event loop.
                return True
            return False

        def gevent_wait_read(self, poll=True):
            if poll and self.gevent_check_read():
                return

            self.__check_watchers()
            wait(self.gevent_read_watcher,
                 hub=self.gevent_hub)

        def gevent_blocking_wait_read(self, poll=True): # pylint:disable=unused-argument
            # poll is for compatibility with ``gevent_wait_read``.
            # Note that there's no timeout here, other than what might be on the
            # socket.
            # XXX: Should there be?
            blocking_select([self], (), ())

        def gevent_generic_wait_read(self, allow_switch=True, poll=True):
            """
            *poll* only applies if *allow_switch* is True.
            """
            waiter = self.gevent_wait_read if allow_switch else self.gevent_blocking_wait_read
            waiter(poll)

        def gevent_check_write(self):
            if gselect((), [self], (), 0)[1]:
                return True
            return False

        def gevent_wait_write(self, poll=True):
            if poll and self.gevent_check_write():
                return

            self.__check_watchers()
            wait(self.gevent_write_watcher,
                 hub=self.gevent_hub)

        def gevent_blocking_wait_write(self, poll=True): # pylint:disable=unused-argument
            blocking_select((), [self], ())

        def gevent_generic_wait_write(self, allow_switch=True, poll=True):
            waiter = self.gevent_wait_write if allow_switch else self.gevent_blocking_wait_write
            waiter(poll)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Base class for ``IRelStorageAdapter``.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time

from persistent.timestamp import TimeStamp

from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64

from relstorage.storage.interfaces import VoteReadConflictError
from .._compat import metricmethod_sampled
from .._compat import MAX_S_TID
from .._util import timestamp_at_unixtime
from .._util import TRACE
from ..options import Options

from ._util import DatabaseHelpersMixin
from .drivers import _select_driver
from .interfaces import UnableToLockRowsToModifyError
from .interfaces import UnableToLockRowsToReadCurrentError

logger = __import__('logging').getLogger(__name__)

class AbstractAdapter(DatabaseHelpersMixin):

    keep_history = None # type: bool
    options = None # type: Options
    driver_options = None # type: IDBDriverOptions
    locker = None # type: ILocker
    txncontrol = None # type: ITransactionControl
    mover = None # type: IObjectMover
    connmanager = None # type: IConnectionManager
    oidallocator = None # type: IOIDAllocator
    dbiter = None # type: DatabaseIterator
    packundo = None

    def __init__(self, options=None):
        if options is None:
            options = Options()
        self.options = options
        self.keep_history = options.keep_history

        self.driver = driver = self._select_driver()
        self._binary = driver.Binary

        # If it was already set, that means it shared with other
        # instances, so no need to register the openings.
        connmanager_was_set = self.connmanager is not None
        self._create()
        if not driver.supports_64bit_unsigned_id:
            self.packundo.MAX_TID = MAX_S_TID
            self.MAX_TID = MAX_S_TID
            self.dbiter.MAX_TID = MAX_S_TID

        if not connmanager_was_set:
            self.connmanager.add_on_store_opened(self.mover.on_store_opened)
            self.connmanager.add_on_load_opened(self.mover.on_load_opened)
            self.connmanager.add_on_store_opened(self.locker.on_store_opened)

    def _create(self):
        raise NotImplementedError

    def release(self):
        if self.oidallocator is not None:
            self.oidallocator.release()
            self.oidallocator = None

    def close(self):
        if self.oidallocator is not None:
            self.oidallocator.close()
            self.oidallocator = None

    def _select_driver(self, options=None):
        return _select_driver(
            options or self.options or Options(),
            self.driver_options
        )

    def __repr__(self):
        return "<%s.%s at 0x%x keep_history=%s driver=%s>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            id(self),
            self.keep_history,
            self.driver,
        )

    @metricmethod_sampled
    def lock_database_and_choose_next_tid(self, cursor,
                                          username,
                                          description,
                                          extension):
        self.locker.hold_commit_lock(cursor, ensure_current=True)

        # Choose a transaction ID.
        #
        # Base the transaction ID on the current time, but ensure that
        # the tid of this transaction is greater than any existing
        # tid.
        last_tid = self.txncontrol.get_tid(cursor)
        now = time.time()
        stamp = timestamp_at_unixtime(now)
        stamp = stamp.laterThan(TimeStamp(int64_to_8bytes(last_tid)))
        tid = stamp.raw()

        tid_int = bytes8_to_int64(tid)
        self.txncontrol.add_transaction(cursor, tid_int, username, description, extension)
        logger.log(TRACE, "Picked next tid locally: %s", tid_int)
        return tid_int

    @metricmethod_sampled
    def lock_database_and_move(self,
                               store_connection, load_connection,
                               transaction_has_blobs,
                               ude,
                               commit=True,
                               committing_tid_int=None,
                               after_selecting_tid=lambda tid: None):
        # Here's where we take the global commit lock, and
        # allocate the next available transaction id, storing it
        # into history-preserving DBs. But if someone passed us
        # a TID (``restore``), then it must already be in the DB, and the lock must
        # already be held.
        #
        # If we've prepared the transaction, then the TID must be in the
        # db, the lock must be held, and we must have finished all of our
        # storage actions. This is only expected to be the case when we have
        # a shared blob dir.

        cursor = store_connection.cursor
        if committing_tid_int is None:
            committing_tid_int = self.lock_database_and_choose_next_tid(
                cursor,
                *ude
            )

        # Move the new states into the permanent table
        # TODO: Figure out how to do as much as possible of this before holding
        # the commit lock. For example, use a dummy TID that we later replace.
        # (This has FK issues in HP dbs).

        self.mover.move_from_temp(cursor, committing_tid_int, transaction_has_blobs)

        after_selecting_tid(committing_tid_int)

        self.mover.update_current(cursor, committing_tid_int)
        prepared_txn_id = self.txncontrol.commit_phase1(
            store_connection, committing_tid_int)

        if commit:
            self.txncontrol.commit_phase2(store_connection, prepared_txn_id, load_connection)

        return committing_tid_int, prepared_txn_id

    DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = True
    WRITING_REQUIRES_EXCLUSIVE_LOCK = False

    # Hooks for unit tests.
    force_lock_objects_and_detect_conflicts_interleavable = False
    force_lock_readCurrent_for_share_blocking = False


    @metricmethod_sampled
    def lock_objects_and_detect_conflicts(self, cursor, read_current_oids):
        if (
                self.force_lock_readCurrent_for_share_blocking
                or self.force_lock_objects_and_detect_conflicts_interleavable
        ):
            # Delegate to the individual statements that can control lock timeouts,
            # or that allow a controlling test to carefully interleave operations to simulate
            # various concurrency situations.
            return self._composed_lock_objects_and_detect_conflicts(cursor,
                                                                    read_current_oids)
        begin = time.time()
        try:
            return self._best_lock_objects_and_detect_conflicts(cursor, read_current_oids)
        except self.locker.lock_exceptions:
            # Heuristic to guess. If the stored proc or stored proc runner can do better,
            # they should.
            elapsed = time.time() - begin
            kind = UnableToLockRowsToModifyError
            if read_current_oids and elapsed < self.locker.commit_lock_timeout:
                kind = UnableToLockRowsToReadCurrentError

            self.locker.reraise_commit_lock_error(
                cursor,
                self._describe_best_lock_objects_and_detect_conflicts(),
                kind
            )

    def _composed_lock_objects_and_detect_conflicts(self, cursor, read_current_oids):
        read_current_oid_ints = read_current_oids.keys()

        def after_lock_share():
            current = self.mover.current_object_tids(cursor, read_current_oid_ints)
            # We go ahead and compare the readCurrent TIDs here, so
            # that we don't have to make the call to detect conflicts
            # or even lock rows if there are readCurrent violations.
            for oid_int, expect_tid_int in read_current_oids.items():
                actual_tid_int = current.get(oid_int, 0)
                if actual_tid_int != expect_tid_int:
                    raise VoteReadConflictError(
                        oid=int64_to_8bytes(oid_int),
                        serials=(int64_to_8bytes(actual_tid_int),
                                 int64_to_8bytes(expect_tid_int)))

        self.locker.lock_current_objects(
            cursor, read_current_oid_ints,
            self.force_lock_readCurrent_for_share_blocking,
            after_lock_share)

        conflicts = self.mover.detect_conflict(cursor)
        return conflicts

    #: Subclasses that have the ability to implement
    #: :meth:`lock_objects_and_detect_conflicts` in a single database
    #: call, or otherwise do better than our
    #: :meth:`_composed_lock_objects_and_detect_conflicts` implementation,
    #: should override this method. It *must* return a materialized list
    #: of conflicts that supports len() and iterating multiple times.
    _best_lock_objects_and_detect_conflicts = _composed_lock_objects_and_detect_conflicts

    def _describe_best_lock_objects_and_detect_conflicts(self):
        return '<unknown>'



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of relstorage.adapters.postgresql"""
from __future__ import absolute_import

import logging
import time
import unittest

from ZODB.tests import StorageTestBase

from relstorage._util import timestamp_at_unixtime
from relstorage.storage import bytes8_to_int64
from relstorage.adapters.postgresql import PostgreSQLAdapter

from .util import AbstractTestSuiteBuilder
from .util import DEFAULT_DATABASE_SERVER_HOST
from . import StorageCreatingMixin
from . import TestCase

class PostgreSQLAdapterMixin(object):

    def make_adapter(self, options, db=None):
        return PostgreSQLAdapter(
            dsn=self.__get_adapter_zconfig_dsn(db),
            options=options,
        )

    def get_adapter_class(self):
        return PostgreSQLAdapter

    def __get_adapter_zconfig_dsn(self, dbname=None):
        if dbname is None:
            if self.keep_history:
                dbname = self.base_dbname
            else:
                dbname = self.base_dbname + '_hf'
        dsn = (
            "dbname='%s' user='relstoragetest' password='relstoragetest' host='%s'"
            % (dbname, DEFAULT_DATABASE_SERVER_HOST)
        )
        return dsn

    def get_adapter_zconfig(self):
        return u"""
        <postgresql>
            driver %s
            dsn %s
        </postgresql>
        """ % (
            self.driver_name,
            self.__get_adapter_zconfig_dsn()
        )

    def verify_adapter_from_zconfig(self, adapter):
        self.assertEqual(adapter._dsn, self.__get_adapter_zconfig_dsn())

class TestBlobMerge(PostgreSQLAdapterMixin,
                    StorageCreatingMixin,
                    TestCase,
                    StorageTestBase.StorageTestBase):
    # pylint:disable=too-many-ancestors

    def test_merge_blobs_on_open(self):
        from ZODB.DB import DB
        from ZODB.blob import Blob
        import transaction
        storage = self._closing(self.make_storage(
            blob_dir='blobs', shared_blob_dir=False))
        db = self._closing(DB(storage))
        conn = db.open()

        blob = Blob()
        base_chunk = b"This is my base blob."
        with blob.open('w') as f:
            f.write(base_chunk)

        conn.root().blob = blob
        transaction.commit()

        # Insert some extra chunks. Get them big to be sure we loop
        # properly
        second_chunk = b'second chunk' * 800
        with conn._storage._store_connection_pool.borrowing(commit=True) as store_connection:
            cursor = store_connection.cursor

            cursor.execute("""
            INSERT INTO blob_chunk (zoid, chunk_num, tid, chunk)
            SELECT zoid, 1, tid, lo_from_bytea(0, %s)
            FROM blob_chunk WHERE chunk_num = 0;
            """, (second_chunk,))
            third_chunk = b'third chunk' * 900
            cursor.execute("""
            INSERT INTO blob_chunk (zoid, chunk_num, tid, chunk)
            SELECT zoid, 2, tid, lo_from_bytea(0, %s)
            FROM blob_chunk WHERE chunk_num = 0;
            """, (third_chunk,))

            cursor.execute('SELECT COUNT(*) FROM blob_chunk')
            self.assertEqual(3, cursor.fetchone()[0])

        # Now open again and find everything put together.
        # But we need to use a new blob dir, because
        # we changed data behind its back.
        conn.close()
        db.close()

        storage = self._closing(self.make_storage(blob_dir='blobs2',
                                                  shared_blob_dir=False,
                                                  zap=False))
        db = self._closing(DB(storage))
        conn = db.open()

        blob = conn.root().blob
        with blob.open('r') as f:
            data = f.read()

        cursor = conn._storage._load_connection.cursor
        cursor.execute('SELECT COUNT(*) FROM blob_chunk')
        self.assertEqual(1, cursor.fetchone()[0])

        self.assertEqual(data, base_chunk + second_chunk + third_chunk)
        conn.close()
        db.close()

class TestGenerateTIDPG(PostgreSQLAdapterMixin,
                        StorageCreatingMixin,
                        TestCase,
                        StorageTestBase.StorageTestBase):
    # pylint:disable=too-many-ancestors

    def setUp(self):
        super(TestGenerateTIDPG, self).setUp()
        self._storage = self._closing(self.make_storage())

    def test_extract_parts(self):
        unix_time = 1564063129.1277142

        query = """
        SELECT EXTRACT(year FROM ts) as YEAR,
               EXTRACT(month FROM ts) AS month,
               EXTRACT(day FROM ts) AS day,
               EXTRACT(hour FROM ts) AS hour,
               EXTRACT(minute FROM ts) AS minute,
               EXTRACT(seconds FROM ts) AS seconds
        FROM (
            SELECT timezone('UTC', to_timestamp(%s)) AS ts
        ) t
        """
        cursor = self._storage._load_connection.cursor
        cursor.execute(query, (unix_time,))
        year, month, day, hour, minute, seconds = cursor.fetchone()
        self.assertEqual(year, 2019)
        self.assertEqual(month, 7)
        self.assertEqual(day, 25)
        self.assertEqual(hour, 13) # If this is not 13, the time_zone is incorrect
        self.assertEqual(minute, 58)
        self.assertEqual(
            round(float(seconds), 6),
            49.127714)

    def test_known_time(self):
        now = 1564054182.277615
        gmtime = (2019, 7, 25, 11, 29, 42, 3, 206, 0)

        self.assertEqual(
            time.gmtime(now),
            gmtime
        )

        ts_now = timestamp_at_unixtime(now)
        self.assertEqual(
            ts_now.raw(),
            b'\x03\xd1Oq\xb4bn\x00'
        )

        self.test_current_time(now)

    def test_current_time(self, now=None):
        if now is None:
            now = time.time()
        storage = self._storage
        ts_now = timestamp_at_unixtime(now)

        expected_tid_int = bytes8_to_int64(ts_now.raw())

        cursor = storage._load_connection.cursor

        cursor.execute('SELECT make_tid_for_epoch(%s)', (now,))
        tid, = cursor.fetchall()[0]

        self.assertEqual(
            tid,
            expected_tid_int
        )

# Timing shows that we spend 6.9s opening database connections to a
# local PostgreSQL 11 server when using Python 3.7 and psycopg2 2.8
# during a total test run of 2:27. I had thought that maybe connection
# pooling would speed the test run up, but that doesn't seem to be the
# case.

class PostgreSQLTestSuiteBuilder(AbstractTestSuiteBuilder):

    __name__ = 'PostgreSQL'

    def __init__(self):
        from relstorage.adapters.postgresql import drivers
        super(PostgreSQLTestSuiteBuilder, self).__init__(
            drivers,
            PostgreSQLAdapterMixin,
            extra_test_classes=(TestBlobMerge, TestGenerateTIDPG)
        )

    def _compute_large_blob_size(self, use_small_blobs):
        if use_small_blobs:
            # Avoid creating 2GB blobs to be friendly to neighbors
            # and to run fast (2GB blobs take about 4 minutes on Travis
            # CI as-of June 2016).
            # RS 3.0 no longer needs to chunk.
            large_blob_size = 20 * 1024 * 1024
        else:
            # Something bigger than 2GB (signed 32 bit int)
            large_blob_size = (1 << 31) + 200 * 1024 * 1024
        return large_blob_size


def test_suite():
    return PostgreSQLTestSuiteBuilder().test_suite()


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger("zc.lockfile").setLevel(logging.CRITICAL)
    unittest.main(defaultTest="test_suite")



# -*- coding: utf-8 -*-
"""
RDBMS-specific SQL.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from contextlib import contextmanager
from operator import attrgetter

from zope.interface import implementer

from relstorage._compat import NStringIO
from relstorage._compat import intern
from ..interfaces import IDBDialect

from .types import OID
from .types import TID
from .types import BinaryString
from .types import State
from .types import Boolean

from ._util import copy
from .interfaces import ITypedParams

# pylint:disable=too-many-function-args

@implementer(IDBDialect)
class DefaultDialect(object):

    keep_history = True
    _context_repr = None

    datatype_map = {
        OID: 'BIGINT',
        TID: 'BIGINT',
        BinaryString: 'BYTEA',
        State: 'BYTEA',
        Boolean: 'BOOLEAN',
    }

    STMT_TRUNCATE = 'TRUNCATE TABLE'
    STMT_TABLE_TRAILER = ''
    STMT_IF_NOT_EXISTS = ''

    def bind(self, context):
        # The context will reference us most likely
        # (compiled statement in instance dictionary)
        # so try to avoid reference cycles.
        keep_history = context.keep_history
        new = copy(self)
        new.keep_history = keep_history
        new._context_repr = repr(context)
        return new

    def compiler_class(self):
        return Compiler

    def compiler(self, root):
        return self.compiler_class()(root, self)

    def datatypes_for_columns(self, column_list):
        columns = list(column_list)
        datatypes = []
        for column in columns:
            datatype = self.datatype_for_column(column)
            datatypes.append(datatype)
        return datatypes

    def datatype_for_column(self, column):
        __traceback_info__ = column
        return column.type_.to_sql_datatype(self.datatype_map).format(
            col_name=column.name
        )

    def extra_constraints_for_column(self, column): # pylint:disable=unused-argument
        return ''

    def __eq__(self, other):
        if isinstance(other, DefaultDialect):
            return other.keep_history == self.keep_history
        return NotImplemented # pragma: no cover

    def __repr__(self):
        return "<%s at %x keep_history=%s context=%s>" % (
            type(self).__name__,
            id(self),
            self.keep_history,
            self._context_repr
        )


class _MissingDialect(DefaultDialect):
    def __bool__(self):
        return False

    __nonzero__ = __bool__


class Compiler(object):
    # pylint:disable=too-many-public-methods

    def __init__(self, root, dialect):
        self.buf = NStringIO()
        self.placeholders = {}
        self.root = root
        self.dialect = dialect # type: DefaultDialect
        self._visit_name_stack = ('__compile_visit__',)

    @contextmanager
    def using_visit_name(self, name):
        if name:
            name = '__compile_visit_for_' + name + '__'
        else:
            name = '__compile_visit__'
        cur_stack = self._visit_name_stack
        self._visit_name_stack = (name,) + self._visit_name_stack
        try:
            yield self
        finally:
            self._visit_name_stack = cur_stack

    def __repr__(self):
        return "<%s %s %r>" % (
            type(self).__name__,
            self.buf.getvalue(),
            self.placeholders
        )

    def compile(self):
        self.visit(self.root)
        return self.finalize()

    def can_prepare(self):
        # Obviously this needs to be more general.
        # Some drivers, for example, can't deal with parameters
        # in a prepared statement; we currently handle that by overriding
        # this method.
        return self.root.prepare

    _prepared_stmt_counter = 0

    def _next_prepared_stmt_name(self, query):
        # Even with the GIL, this isn't fully safe to do; two threads
        # can still get the same value. We don't want to allocate a
        # lock because we might be patched by gevent later. So that's
        # where `query` comes in: we add the hash as a disambiguator.
        # Of course, for there to be a duplicate prepared statement
        # sent to the database, that would mean that we were somehow
        # using the same cursor or connection in multiple threads at
        # once (or perhaps we got more than one cursor from a
        # connection? We should only have one.)
        #
        # TODO: Sidestep this problem by allocating this earlier;
        # the SELECT or INSERT statement could pick it when it is created;
        # that happens at the class level at import time, when we should be
        # single-threaded.
        #
        # That may also help facilitate caching.
        Compiler._prepared_stmt_counter += 1
        return 'rs_prep_stmt_%s_%d_%d' % (
            getattr(self.root, "__name__", ''),
            Compiler._prepared_stmt_counter,
            abs(hash(query)),
        )

    def _prepared_param(self, number):
        return '$' + str(number)

    _PREPARED_CONJUNCTION = 'AS'

    def _quote_query_for_prepare(self, query):
        return query

    def _find_datatypes_for_prepared_query(self):
        # Deduce the datatypes based on the types of the columns
        # we're sending as params.
        root = self.root
        if ITypedParams.providedBy(root):
            # Don't call the interface to adapt; that uses adapter
            # hooks, which may be persistent objects and thus result
            # in calling back into this method (if the query we're
            # compiling for the first time is the load object query,
            # and the site manager itself is a ghost).
            #
            # Since we don't actually expect to have any adapters,
            # just an object (Insert or a subclass) that already
            # provides the interface, all we need to do is check to
            # see if it's provided; we don't need to query an adapter.
            # See https://github.com/zodb/relstorage/issues/411
            return root.datatypes_for_parameters()
        return ()

    def prepare(self):
        # This is correct for PostgreSQL. This needs moved to a dialect specific
        # spot.

        datatypes = self._find_datatypes_for_prepared_query()
        query = self.buf.getvalue()
        name = self._next_prepared_stmt_name(query)

        if datatypes:
            assert isinstance(datatypes, (list, tuple))
            datatypes = ', '.join(datatypes)
            datatypes = ' (%s)' % (datatypes,)
        else:
            datatypes = ''

        q = query.strip()

        # PREPARE needs the query string to use $1, $2, $3, etc,
        # as placeholders.
        # In MySQL, it's a plain question mark.
        placeholder_to_number = {}
        counter = 0
        for placeholder_name in self.placeholders.values():
            counter += 1
            placeholder = self._placeholder(placeholder_name)
            placeholder_to_number[placeholder_name] = counter
            param = self._prepared_param(counter)
            q = q.replace(placeholder, param, 1)

        q = self._quote_query_for_prepare(q)

        stmt = 'PREPARE {name}{datatypes} {conjunction} {query}'.format(
            name=name, datatypes=datatypes,
            query=q,
            conjunction=self._PREPARED_CONJUNCTION,
        )


        if placeholder_to_number:
            execute = 'EXECUTE {name}({params})'.format(
                name=name,
                params=','.join(['%s'] * len(self.placeholders)),
            )
        else:
            # Neither MySQL nor PostgreSQL like a set of empty parens: ()
            execute = 'EXECUTE {name}'.format(name=name)

        if '%s' in placeholder_to_number:
            # There was an ordered param. If there was one,
            # they must all be ordered, so there's no need to convert anything.
            assert len(placeholder_to_number) == 1
            def convert(p):
                return p
        else:
            def convert(d):
                # TODO: This may not actually be needed, since we issue a regular
                # cursor.execute(), it may be able to handle named?
                params = [None] * len(placeholder_to_number)
                for placeholder_name, ix in placeholder_to_number.items():
                    params[ix - 1] = d[placeholder_name]
                return params

        return intern(stmt), intern(execute), convert

    def finalize(self):
        return intern(self.buf.getvalue().strip()), {v: k for k, v in self.placeholders.items()}

    def visit(self, node, **kwargs):
        """
        Returns whatever the ``__compile_visit__`` method of the *node*
        returns.
        """
        for visit_name in self._visit_name_stack:
            meth = getattr(node, visit_name, None)
            if meth:
                return meth(self, **kwargs)

        raise AttributeError("No way to visit node", node)

    visit_clause = visit
    visit_upsert = visit

    def emit(self, *contents):
        for content in contents:
            self.buf.write(content)

    def emit_null(self):
        self.emit('NULL')

    def emit_w_padding_space(self, value):
        ended_in_space = self.buf.getvalue().endswith(' ')
        value = value.strip()
        if not ended_in_space:
            self.buf.write(' ')
        self.emit(value, ' ')

    emit_keyword = emit_w_padding_space
    emit_column_constraint = emit_w_padding_space

    def emit_column_autoincrement(self):
        self.emit_keyword(self.dialect.CONSTRAINT_AUTO_INCREMENT)

    def emit_keyword_truncate_table(self):
        self.emit_keyword(self.dialect.STMT_TRUNCATE)

    def emit_keyword_upsert(self):
        self.emit_keyword_insert_into()

    def emit_keyword_insert_into(self):
        self.emit_keyword("INSERT INTO")

    def emit_identifier(self, identifier, quoted=False):
        last_char = self.buf.getvalue()[-1]

        if quoted:
            emit = ('"', identifier, '"')
        else:
            emit = (identifier,)
        if last_char not in ('(', ' '):
            emit = (' ',) + emit

        self.emit(*emit)

    def emit_quoted_identifier(self, identifier):
        self.emit_identifier(identifier, quoted=True)

    def visit_select_list(self, column_list):
        clist = column_list.c if hasattr(column_list, 'c') else column_list
        self.visit(clist.as_select_list())

    def visit_csv(self, nodes):
        self.visit(nodes[0])
        for node in nodes[1:]:
            self.emit(', ')
            self.visit(node)

    visit_select_expression = visit

    def visit_select_list_csv(self, nodes):
        self.visit_select_expression(nodes[0])
        for node in nodes[1:]:
            self.emit(', ')
            self.visit_select_expression(node)

    def visit_column(self, column_node):
        self.emit_identifier(column_node.name)

    def _visit_or_str(self, node_or_name):
        if hasattr(node_or_name, '__compile_visit__'):
            self.visit(node_or_name)
        else:
            self.emit_identifier(node_or_name)

    def visit_aliased_column(self, column_node):
        self._visit_or_str(column_node.name)
        self.emit_identifier(column_node.alias)

    def visit_qualified_column(self, column):
        self.emit_identifier(column.table + '.' + column.name)

    def visit_from(self, from_):
        self.emit_keyword('FROM')
        self.visit(from_)

    def visit_grouped(self, clause):
        self.emit('(')
        self.visit(clause)
        self.emit(')')

    def visit_no_values(self):
        self.emit('()')

    def visit_op(self, op):
        last_char = self.buf.getvalue()[-1]
        if last_char != ' ':
            self.emit(' ')
        self.emit(op, ' ')

    def _next_placeholder_name(self, prefix='param'):
        return '%s_%d' % (prefix, len(self.placeholders),)

    def _placeholder(self, key):
        # Write things in `pyformat` style by default, assuming a
        # dictionary of params; this is supported by most drivers.
        if key == '%s':
            return key
        return '%%(%s)s' % (key,)

    def _placeholder_for_literal_param_value(self, value):
        placeholder = self.placeholders.get(value)
        if not placeholder:
            placeholder_name = self._next_placeholder_name(prefix='literal')
            placeholder = self._placeholder(placeholder_name)
            self.placeholders[value] = placeholder_name
        return placeholder

    def visit_literal_expression(self, value):
        if isinstance(value, int):
            self.emit(str(value))
        else:
            placeholder = self._placeholder_for_literal_param_value(value)
            self.emit(placeholder)

    def visit_boolean_literal_expression(self, value):
        # In the oracle dialect, this needs to be
        # either "'Y'" or "'N'". sqlite supports only 1 and 0,
        # prior to certain versions.
        assert isinstance(value, bool)
        self.emit(str(value).upper())

    def visit_immediate_expression(self, value):
        # Like a literal, but never uses a placeholder.
        if isinstance(value, bool):
            self.visit_boolean_literal_expression(value)
        else:
            self.emit(str(value))

    def visit_bind_param(self, bind_param):
        self.placeholders[bind_param] = bind_param.key
        self.emit(self._placeholder(bind_param.key))

    def visit_ordered_bind_param(self, bind_param):
        self.placeholders[bind_param] = '%s'
        self.emit('%s')

    def create_table(self, table, if_not_exists, use_trailer=True):
        with self.using_visit_name('create'):
            self.visit(table, if_not_exists=if_not_exists)
            if use_trailer:
                self.emit(' ', self.dialect.STMT_TABLE_TRAILER)

    def emit_if_not_exists(self):
        if self.dialect.STMT_IF_NOT_EXISTS:
            self.emit(self.dialect.STMT_IF_NOT_EXISTS)

    @contextmanager
    def visit_limited_select(self, select, limit): # pylint:disable=unused-argument
        yield self

    def visit_limit(self, limit_literal):
        if limit_literal:
            self.emit_keyword('LIMIT')
            self.visit_immediate_expression(limit_literal)

    @contextmanager
    def visiting_upsert(self, upsert): # pylint:disable=unused-argument
        with self.using_visit_name('upsert'):
            yield self

    def visit_upsert_conflict_column(self, column):
        self.emit_keyword('ON CONFLICT')
        self.visit_grouped(column)

    def visit_upsert_conflict_update(self, update):
        self.emit_keyword('DO')
        self.visit(update)

    def visit_upsert_excluded_column(self, column):
        self.emit('excluded.')
        self.emit(column.name)

    def visit_upsert_after_select(self, select): # pylint:disable=unused-argument
        pass

class _DefaultContext(object):

    keep_history = True


class DialectAware(object):

    context = _DefaultContext()
    dialect = _MissingDialect()

    _dialect_locations = (
        attrgetter('dialect'),
        attrgetter('driver.dialect'),
        attrgetter('poller.driver.dialect'),
        attrgetter('connmanager.driver.dialect'),
        attrgetter('adapter.driver.dialect')
    )

    def _find_dialect(self, context):
        # Find the dialect to use for the context. If it specifies
        # one, then use it. Otherwise go hunting for the database
        # driver and use *it*. Preferably the driver is attached to
        # the object we're looking at, but if not that, we'll look at
        # some common attributes for adapter objects for it.
        if isinstance(context, DefaultDialect):
            return context

        for getter in self._dialect_locations:
            try:
                dialect = getter(context)
            except AttributeError:
                pass
            else:
                return dialect.bind(context)
        __traceback_info__ = getattr(context, '__dict__', ()) # vars() doesn't work on e.g., None
        raise TypeError("Unable to bind to %s; no dialect found" % (context,))

    def bind(self, context, dialect=None):
        assert self.context is DialectAware.context, "already bound"
        if dialect is None:
            dialect = self._find_dialect(context)

        assert dialect is not None

        new = copy(self)
        return new._bound_to(context, dialect)

    _bind_vars_ignored = ()

    def _bound_to(self, context, dialect):
        # Called on the copy of self.
        if context is not None:
            self.context = context
        self.dialect = dialect

        bound_replacements = {
            k: v.bind(context, dialect)
            for k, v
            in vars(self).items()
            if k not in self._bind_vars_ignored and isinstance(v, DialectAware)
        }
        for k, v in bound_replacements.items():
            setattr(self, k, v)

        return self



from __future__ import print_function

import logging
import os
import random

import transaction
from persistent.mapping import PersistentMapping
from ZODB.DB import DB

from relstorage.options import Options
from relstorage.storage import RelStorage

logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)

use = 'oracle'
keep_history = True

if use == 'mysql':
    from relstorage.adapters.mysql import MySQLAdapter
    a = MySQLAdapter(db='packtest',
                     user='relstoragetest',
                     passwd='relstoragetest',
                     options=Options(keep_history=keep_history),)
elif use == 'postgresql':
    from relstorage.adapters.postgresql import PostgreSQLAdapter
    a = PostgreSQLAdapter(dsn="dbname='packtest' "
                          'user=relstoragetest '
                          'password=relstoragetest',
                          options=Options(keep_history=keep_history),)
elif use == 'oracle':
    from relstorage.adapters.oracle import OracleAdapter
    dsn = os.environ.get('ORACLE_TEST_DSN', 'XE')
    a = OracleAdapter(user='packtest',
                      password='relstoragetest',
                      dsn=dsn,
                      options=Options(keep_history=keep_history),)
else:
    raise AssertionError("which database?")

s = RelStorage(a)
d = DB(s)
c = d.open()

print('size:')
print(d.getSize())


print('initializing...')
container = PersistentMapping()
c.root()['container'] = container
container_size = 10000
for i in range(container_size):
    container[i] = PersistentMapping()
transaction.commit()

print('generating transactions...')
for trans in range(100):
    print(trans)
    sources = (random.randint(0, container_size - 1) for j in range(100))
    for source in sources:
        obj = container[source]
        obj[trans] = container[random.randint(0, container_size - 1)]
    transaction.commit()

print('size:')
print(d.getSize())

print('packing...')
d.pack()

print('size:')
print(d.getSize())

d.close()
